import json


def tileset_info(filename):
    """
    Return a tileset info for this time interval
    """
    with open(filename, "r") as f:
        data = json.load(f)

        return {
            "min_pos": [0],
            "max_pos": [data["len"]],
            "max_width": data["len"],
            "start_value": data["start"],
            "end_value": data["end"],
        }



import numpy as np


def aggregate(x, k):
    x = np.asarray(x)

    if len(x) % k != 0:
        extend = k - len(x) % k
        x = np.r_[x, [np.nan] * extend]

    xr = x.reshape((-1, k))
    is_allnan = np.all(np.isnan(xr), axis=1)
    agg = np.nansum(xr, axis=1)

    if np.any(is_allnan):
        if not issubclass(agg.dtype.type, np.floating):
            agg = agg.astype(np.float64)
        agg[is_allnan] = np.nan

    return agg



import clodius.fpark as fp


def test_map():
    a = fp.FakeSparkContext.parallelize([1, 2, 3, 4])

    b = a.map(lambda x: x + 1)
    assert b.collect()[0] == 2


def test_group_by_key():
    a = fp.FakeSparkContext.parallelize([(1, 2), (1, 3), (1, 4), (2, 5), (2, 6)])

    a.groupByKey()
    # TODO: Make assertions about result


def test_textFile():
    a = fp.FakeSparkContext.textFile("test/sample_data/piecewise-file/part-00000")
    parts = a.collect()
    assert len(parts) == 2

    a = fp.FakeSparkContext.textFile("test/sample_data/piecewise-file/")
    parts = a.collect()
    assert len(parts) == 4



from clodius.tiles.format import format_dense_tile
from clodius.tiles.utils import tile_bounds
import pandas as pd

import numpy as np
import h5py


def csv_to_points(csv_file, output_file, max_zoom: int = 30):
    """
    Convert a csv file containing points to a numpy array
    of [[x,y]] values.

    Parameters:
    -----------
    csv_file: string
        The filename of the data file

    """
    df = pd.read_csv(csv_file)

    min_x = df["x"].min()
    max_x = df["x"].max()
    min_y = df["y"].min()
    max_y = df["y"].max()

    width = max_x - min_x
    height = max_y - min_y
    max_width = max(width, height)

    with h5py.File(output_file, "w") as f_out:
        dataset = f_out.create_dataset(
            "values", (len(df), 2), compression="gzip", dtype=np.float32
        )
        dataset[:] = df.reindex(columns=["x", "y"]).values

        dataset.attrs["min_x"] = min_x
        dataset.attrs["max_x"] = max_x
        dataset.attrs["min_y"] = min_y
        dataset.attrs["max_y"] = max_y
        dataset.attrs["max_zoom"] = max_zoom
        dataset.attrs["max_width"] = max_width

    return df.reindex(columns=["x", "y"])


def tileset_info(points_file):
    """
    Calculate the extent, etc...
    """
    with h5py.File(points_file, "r") as f_in:
        attrs = f_in["values"].attrs

        return {
            "min_pos": [float(attrs["min_x"]), float(attrs["min_y"])],
            "max_pos": [float(attrs["max_x"]), float(attrs["max_y"])],
            "max_width": float(attrs["max_width"]),
            "max_zoom": int(attrs["max_zoom"]),
            "mirror_tiles": "false",
        }


def filter_points(data, extent):
    """
    Filter points that are within the extent

    Parameters:
    -----------
    data: [[]]
        A 2D numpy array containing x,y values

    extent: [x_start, x_end, y_start, y_end]
        The region we want to return points within

    Returns
    -------
    data: [[]]
        A 2D numpy array containing x,y values
    """
    # print("extent:", extent)
    # print("data.shape", data.shape, data[:,0])
    data = data[data[:, 0] > extent[0]]
    data = data[data[:, 0] < extent[2]]

    data = data[data[:, 1] > extent[1]]
    data = data[data[:, 1] < extent[3]]

    return data


def density_tiles(points_file, z, x, y, width=1, height=1):
    """
    Get a 2D histogram of the given region. If the height and
    width are specified, then we need to partition this into
    multiple returned tiles.
    """
    returns = []

    with h5py.File(points_file, "r") as f:
        # get all the points in the region
        all_points = filter_points(
            f["values"][:],
            tile_bounds(tileset_info(points_file), z, x, y, width, height),
        )

        for i in range(width):
            for j in range(height):
                # filter from the larger subregion
                filtered_points = filter_points(
                    all_points, tile_bounds(tileset_info(points_file), z, x + i, y + j)
                )

                dt = np.histogram2d(
                    filtered_points[:, 0], filtered_points[:, 1], bins=256
                )[0].T
                dt[dt == 0.0] = np.nan

                returns += [((z, x + i, y + j), dt)]

        return returns


def tiles(points_file, z, x, y, width=1, height=1):
    return [
        (tile_pos, format_dense_tile(data.flatten()))
        for (tile_pos, data) in density_tiles(points_file, z, x, y, width, height)
    ]



# -*- coding: utf-8 -*-
from __future__ import division, print_function

from . import cli

import click
import clodius.chromosomes as cch
import clodius.multivec as cmv
import clodius.array as ct
import collections as col
import h5py
import math
import negspy.coordinates as nc
import numpy as np
import os
import os.path as op
import random
import scipy.misc as sm
import slugid
import sqlite3
import sys
import time
import gzip
import json

from .utils import get_tile_pos_from_lng_lat


@cli.group()
def aggregate():
    """
    Aggregate a data file so that it stores the data at multiple
    resolutions.
    """
    pass


def store_meta_data(
    cursor,
    zoom_step,
    max_length,
    assembly,
    chrom_names,
    chrom_sizes,
    tile_size,
    max_zoom,
    max_width,
    version,
    header=[],
):
    cursor.execute(
        """
        CREATE TABLE tileset_info
        (
            zoom_step INT,
            max_length INT,
            assembly text,
            chrom_names text,
            chrom_sizes text,
            tile_size REAL,
            max_zoom INT,
            max_width REAL,
            header text,
            version text
        )
        """
    )

    cursor.execute(
        "INSERT INTO tileset_info VALUES (?,?,?,?,?,?,?,?,?,?)",
        (
            zoom_step,
            max_length,
            assembly,
            "\t".join(chrom_names),
            "\t".join(map(str, chrom_sizes)),
            tile_size,
            max_zoom,
            max_width,
            "\t".join(header),
            version,
        ),
    )

    cursor.commit()

    pass


# all entries are broked up into ((tile_pos), [entry]) tuples
# we just need to reduce the tiles so that no tile contains more than
# max_entries_per_tile entries
# (notice that [entry] is an array), this format will be important when
# reducing to the most important values
def reduce_values_by_importance(
    entry1, entry2, max_entries_per_tile=100, reverse_importance=False
):
    def extract_key(entries):
        return [(e[-2], e) for e in entries]

    by_uid = dict(extract_key(entry1) + extract_key(entry2))
    combined_by_uid = by_uid.values()

    if reverse_importance:
        combined_entries = sorted(combined_by_uid, key=lambda x: float(x[-1]))
    else:
        combined_entries = sorted(combined_by_uid, key=lambda x: -float(x[-1]))

    return combined_entries[:max_entries_per_tile]


def _multivec(
    filepath,
    output_file,
    assembly,
    tile_size,
    chromsizes_filename,
    starting_resolution,
    method,
    row_infos_filename=None,
):
    """
    Aggregate a multivec file.

    This is a file containing nxn data that is aggregated along only one axis.
    This data should be in an HDF5 file where each dataset is named for a
    chromosome and contains a 'resolutions' group containing values for the
    base level resolution.

    Example: f['chr1']['reslutions']['1000'] = [[1,2,3],[4,5,6]]

    The resulting data will be organized by resolution and chromosome.

    Example: f_out['chr1']['resolutions']['5000']=[[1000,2000,3000],[4000,5000,6000]]

    Aggregation is currently done by summing adjacent values.
    """
    f_in = h5py.File(filepath, "r")

    if output_file is None:
        output_file = op.splitext(filepath)[0] + ".multires.mv5"

    (chrom_info, chrom_names, chrom_sizes) = cch.load_chromsizes(
        chromsizes_filename, assembly
    )

    # TODO: "method" is not defined, so this would not work?
    if method == "maxtotal":  # noqa: F821
        pass
    if method == "logsumexp":  # noqa: F821

        def agg(x):
            a = x.T.reshape((x.shape[1], -1, 2))
            return sm.logsumexp(a, axis=2).T

    else:

        def agg(x):
            return x.T.reshape((x.shape[1], -1, 2)).sum(axis=2).T

    print("agg:", agg)
    if row_infos_filename is not None:
        with open(row_infos_filename, "r") as fr:
            row_infos = [l.strip().encode("utf8") for l in fr]
    else:
        row_infos = None
    print("row_infos:", row_infos)

    cmv.create_multivec_multires(
        f_in,
        chromsizes=zip(chrom_names, chrom_sizes),
        agg=lambda x: np.nansum(x.T.reshape((x.shape[1], -1, 2)), axis=2).T,
        starting_resolution=starting_resolution,
        tile_size=tile_size,
        output_file=output_file,
        row_infos=row_infos,
    )


def _bedpe(
    filepath,
    output_file,
    assembly,
    importance_column,
    has_header,
    max_per_tile,
    tile_size,
    chromosome=None,
    chromsizes_filename=None,
    chr1_col=1,
    from1_col=2,
    to1_col=3,
    chr2_col=4,
    from2_col=5,
    to2_col=6,
    max_zoom=None,
):
    BED2DDB_VERSION = 1

    print("output_file:", output_file)

    if filepath == "-":
        f = sys.stdin
    elif filepath.endswith(".gz"):
        f = gzip.open(filepath, "rt")
    else:
        print("plain")
        f = open(filepath, "r")

    if output_file is None:
        output_file = filepath + ".multires.db"
    else:
        output_file = output_file

    if op.exists(output_file):
        os.remove(output_file)

    (chrom_info, chrom_names, chrom_sizes) = cch.load_chromsizes(
        chromsizes_filename, assembly
    )

    def line_to_dict(line):
        parts = line.split()
        d = {}
        try:
            d["xs"] = [
                chrom_info.cum_chrom_lengths[parts[chr1_col - 1]]
                + int(parts[from1_col - 1]),
                chrom_info.cum_chrom_lengths[parts[chr1_col - 1]]
                + int(parts[to1_col - 1]),
            ]
            d["ys"] = [
                chrom_info.cum_chrom_lengths[parts[chr2_col - 1]]
                + int(parts[from2_col - 1]),
                chrom_info.cum_chrom_lengths[parts[chr2_col - 1]]
                + int(parts[to2_col - 1]),
            ]
        except KeyError:
            error_str = (
                "ERROR converting chromosome position to genome position. "
                "Please make sure you've specified the correct assembly "
                "using the --assembly option or a chromsizes file using the . "
                "--chromsizes-filename option."
                "Current assembly: {}, chromosomes: {},{}".format(
                    assembly, parts[chr1_col - 1], parts[chr2_col - 1]
                )
            )
            raise (KeyError(error_str))

        d["uid"] = slugid.nice()

        d["chrOffset"] = d["xs"][0] - int(parts[from1_col - 1])

        if importance_column is None:
            d["importance"] = max(d["xs"][1] - d["xs"][0], d["ys"][1] - d["ys"][0])
        elif importance_column == "random":
            d["importance"] = random.random()
        else:
            # We seem to use one-based numbering for columns...
            d["importance"] = float(parts[int(importance_column) - 1])

        d["fields"] = line

        return d

    entries = []

    if has_header:
        f.readline()
    else:
        first_line = f.readline().strip()
        try:
            parts = first_line.split()

            """
            print("chr1_col", chr1_col, "chr2_col", chr2_col,
                  "from1_col:", from1_col, "from2_col", from2_col,
                  "to1_col", to1_col, "to2_col", to2_col)
            """

            int(parts[from1_col - 1])
            int(parts[to1_col - 1])
            int(parts[from2_col - 1])
            int(parts[to2_col - 1])
        except ValueError:
            error_str = (
                "Couldn't convert one of the bedpe coordinates to an "
                "integer. If the input file contains a header, make sure to "
                "indicate that with the --has-header option. Line: {}".format(
                    first_line
                )
            )
            raise ValueError(error_str)
        entries = [line_to_dict(first_line)]

    entries += [line_to_dict(line) for line in [line.strip() for line in f] if line]

    # We neeed chromosome information as well as the assembly size to properly
    # tile this data
    tile_size = tile_size
    assembly_size = chrom_info.total_length + 1
    max_zoom = int(math.ceil(math.log(assembly_size / tile_size) / math.log(2)))
    """
    if max_zoom is not None and max_zoom < max_zoom:
        max_zoom = max_zoom
    """

    # this script stores data in a sqlite database
    sqlite3.register_adapter(np.int64, lambda val: int(val))
    conn = sqlite3.connect(output_file)

    # store some meta data
    store_meta_data(
        conn,
        1,
        max_length=assembly_size,
        assembly=assembly,
        chrom_names=chrom_names,
        chrom_sizes=chrom_sizes,
        tile_size=tile_size,
        max_zoom=max_zoom,
        max_width=tile_size * 2 ** max_zoom,
        version=BED2DDB_VERSION,
    )

    # max_width = tile_size * 2 ** max_zoom
    # uid_to_entry = {}

    c = conn.cursor()
    c.execute(
        """
        CREATE TABLE intervals
        (
            id int PRIMARY KEY,
            zoomLevel int,
            importance real,
            fromX int,
            toX int,
            fromY int,
            toY int,
            chrOffset int,
            uid text,
            fields text
        )
        """
    )

    c.execute(
        """
        CREATE VIRTUAL TABLE position_index USING rtree(
            id,
            rFromX, rToX,
            rFromY, rToY
        )
        """
    )

    curr_zoom = 0
    counter = 0

    tile_counts = col.defaultdict(lambda: col.defaultdict(lambda: col.defaultdict(int)))
    entries = sorted(entries, key=lambda x: -x["importance"])

    counter = 0
    for d in entries:
        curr_zoom = 0

        while curr_zoom <= max_zoom:
            tile_width = tile_size * 2 ** (max_zoom - curr_zoom)
            tile_from = list(map(lambda x: x / tile_width, [d["xs"][0], d["ys"][0]]))
            tile_to = list(map(lambda x: x / tile_width, [d["xs"][1], d["ys"][1]]))

            empty_tiles = True

            # go through and check if any of the tiles at this zoom level are
            # full

            for i in range(int(tile_from[0]), int(tile_to[0]) + 1):
                if not empty_tiles:
                    break

                for j in range(int(tile_from[1]), int(tile_to[1]) + 1):
                    if tile_counts[curr_zoom][i][j] > max_per_tile:

                        empty_tiles = False
                        break

            if empty_tiles:
                # they're all empty so add this interval to this zoom level
                for i in range(int(tile_from[0]), int(tile_to[0]) + 1):
                    for j in range(int(tile_from[1]), int(tile_to[1]) + 1):
                        tile_counts[curr_zoom][i][j] += 1

                c.execute(
                    "INSERT INTO intervals VALUES (?,?,?,?,?,?,?,?,?,?)",
                    (
                        counter,
                        curr_zoom,
                        d["importance"],
                        d["xs"][0],
                        d["xs"][1],
                        d["ys"][0],
                        d["ys"][1],
                        d["chrOffset"],
                        d["uid"],
                        d["fields"],
                    ),
                )
                conn.commit()

                c.execute(
                    "INSERT INTO position_index VALUES (?,?,?,?,?)",
                    (
                        counter,
                        d["xs"][0],
                        d["xs"][1],
                        d["ys"][0],
                        d["ys"][1],
                    ),  # add counter as a primary key
                )
                conn.commit()

                counter += 1
                break

            curr_zoom += 1

    return


def _bedfile(
    filepath,
    output_file,
    assembly,
    importance_column,
    has_header,
    chromosome,
    max_per_tile,
    tile_size,
    delimiter,
    chromsizes_filename,
    offset,
):
    BEDDB_VERSION = 3

    if output_file is None:
        output_file = filepath + ".beddb"
    else:
        output_file = output_file

    if op.exists(output_file):
        os.remove(output_file)

    if filepath.endswith(".gz"):
        import gzip

        bed_file = gzip.open(filepath, "rt")
    else:
        bed_file = open(filepath, "r")

    try:
        (chrom_info, chrom_names, chrom_sizes) = cch.load_chromsizes(
            chromsizes_filename, assembly
        )
    except FileNotFoundError:
        if chromsizes_filename is None:
            print("Assembly not found:", assembly, file=sys.stderr)
        else:
            print(
                "Chromsizes filename not found:", chromsizes_filename, file=sys.stderr
            )
        return None

    rand = random.Random(3)

    def line_to_np_array(line):
        """
        Convert a bed file line to a numpy array which can later
        be used as an entry in an h5py file.
        """
        try:
            start = int(line[1])
            stop = int(line[2])
        except ValueError:
            raise ValueError("Error parsing the position, line: {}".format(line))

        chrom = line[0]

        if importance_column is None:
            # assume a random importance when no aggregation strategy is given
            importance = rand.random()
        elif importance_column == "size":
            importance = stop - start
        elif importance_column == "random":
            importance = rand.random()
        else:
            importance = float(line[int(importance_column) - 1])

        if stop < start:
            print("WARNING: stop < start:", line, file=sys.stderr)

            start, stop = stop, start

        if len(line) > 3:
            bedline_name = line[3]
        else:
            bedline_name = ""
        # convert chromosome coordinates to genome coordinates
        genome_start = chrom_info.cum_chrom_lengths[chrom] + start + offset
        genome_end = chrom_info.cum_chrom_lengths[chrom] + stop + offset

        pos_offset = genome_start - start
        parts = {
            "startPos": genome_start,
            "endPos": genome_end,
            "uid": slugid.nice(),
            "name": bedline_name,
            "chrOffset": pos_offset,
            "fields": "\t".join(line),
            "importance": importance,
            "chromosome": str(chrom),
        }

        return parts

    dset = []

    print("delimiter:", delimiter)
    if has_header:
        line = bed_file.readline()
        header = line.strip().split(delimiter)
    else:
        line = bed_file.readline().strip()
        line_parts = line.strip().split(delimiter)
        try:
            dset += [line_to_np_array(line_parts)]
        except KeyError:
            print(
                f"Unable to find {line_parts[0]} in the list of chromosome sizes. "
                "Please make sure the correct assembly or chromsizes filename "
                "is passed in as a parameter",
                file=sys.stderr,
            )
            return None
        except IndexError:
            print("Invalid line:", line)
        header = map(str, list(range(1, len(line.strip().split(delimiter)) + 1)))

    for line in bed_file:
        line_parts = line.strip().split(delimiter)
        try:
            dset += [line_to_np_array(line_parts)]
        except IndexError:
            print("Invalid line:", line)

    if chromosome is not None:
        dset = [d for d in dset if d["chromosome"] == chromosome]

    # We neeed chromosome information as well as the assembly size to properly
    # tile this data
    tile_size = tile_size

    assembly_size = chrom_info.total_length + 1
    """
    else:
        try:
            assembly_size = chrom_info.chrom_lengths[chromosome]
        except KeyError:
            print(
                "ERROR: Chromosome {} not found in assembly {}.".format(
                    chromosome, assembly
                ),
                file=sys.stderr
            )
            return 1
    """

    max_zoom = int(math.ceil(math.log(assembly_size / tile_size) / math.log(2)))
    """
    if max_zoom is not None and max_zoom < max_zoom:
        max_zoom = max_zoom
    """

    # this script stores data in a sqlite database
    import sqlite3

    sqlite3.register_adapter(np.int64, lambda val: int(val))
    print("output_file:", output_file, "header:", header)
    conn = sqlite3.connect(output_file)

    # store some meta data
    store_meta_data(
        conn,
        1,
        max_length=assembly_size,
        assembly=assembly,
        chrom_names=chrom_names,
        chrom_sizes=chrom_sizes,
        tile_size=tile_size,
        max_zoom=max_zoom,
        max_width=tile_size * 2 ** max_zoom,
        header=header,
        version=BEDDB_VERSION,
    )

    # max_width = tile_size * 2 ** max_zoom
    uid_to_entry = {}

    intervals = []

    # store each bed file entry as an interval
    for d in dset:
        uid = d["uid"]
        uid_to_entry[uid] = d
        intervals += [(d["startPos"], d["endPos"], uid)]

    tile_width = tile_size

    c = conn.cursor()
    c.execute(
        """
        CREATE TABLE intervals
        (
            id int PRIMARY KEY,
            zoomLevel int,
            importance real,
            startPos int,
            endPos int,
            chrOffset int,
            uid text,
            name text,
            fields text
        )
        """
    )

    c.execute(
        """
        CREATE VIRTUAL TABLE position_index USING rtree(
            id,
            rStartZoomLevel, rEndZoomLevel, rStartPos, rEndPos
        )
        """
    )

    curr_zoom = 0
    counter = 0

    max_viewable_zoom = max_zoom

    if max_zoom is not None and max_zoom < max_zoom:
        max_viewable_zoom = max_zoom

    sorted_intervals = sorted(
        intervals, key=lambda x: -uid_to_entry[x[-1]]["importance"]
    )
    # print('si:', sorted_intervals[:10])
    print("max_per_tile:", max_per_tile)

    tile_counts = col.defaultdict(int)

    for interval in sorted_intervals:
        # go through each interval from most important to least
        while curr_zoom <= max_viewable_zoom:
            # try to place it in the highest zoom level and go down from there
            tile_width = tile_size * 2 ** (max_zoom - curr_zoom)

            curr_pos = interval[0]
            space_available = True

            # check if there's space at this zoom level
            while curr_pos < interval[1]:
                curr_tile = math.floor(curr_pos / tile_width)
                tile_id = "{}.{}".format(curr_zoom, curr_tile)

                """
                if interval[0] < 1000000:
                    print('tile_id:', tile_id, tile_counts[tile_id], curr_zoom, 'interval:', interval)
                """

                # print(tile_id, "tile_counts[tile_id]", tile_counts[tile_id])
                if tile_counts[tile_id] >= max_per_tile:
                    space_available = False
                    break

                curr_pos += tile_width

            # if there is, then fill it up
            if space_available:
                curr_pos = interval[0]
                while curr_pos < interval[1]:
                    curr_tile = math.floor(curr_pos / tile_width)
                    tile_id = "{}.{}".format(curr_zoom, curr_tile)

                    tile_counts[tile_id] += 1

                    """
                    # increment tile counts for lower level tiles
                    higher_zoom = curr_zoom + 1
                    higher_tile = math.floor(higher_zoom / 2)

                    while higher_zoom <= max_viewable_zoom:
                        new_tile_id = '{}.{}'.format(higher_zoom, higher_tile)
                        higher_zoom += 1
                        higher_tile = math.floor(higher_tile / 2)
                        tile_counts[new_tile_id] += 1
                    """

                    curr_pos += tile_width

            if space_available:
                # there's available space
                value = uid_to_entry[interval[-1]]

                # one extra question mark for the primary key
                exec_statement = "INSERT INTO intervals VALUES (?,?,?,?,?,?,?,?,?)"

                c.execute(
                    exec_statement,
                    # primary key, zoomLevel, startPos, endPos, chrOffset, line
                    (
                        counter,
                        curr_zoom,
                        value["importance"],
                        value["startPos"],
                        value["endPos"],
                        value["chrOffset"],
                        value["uid"],
                        value["name"],
                        value["fields"],
                    ),
                )

                if counter % 1000 == 0:
                    print("counter:", counter, value["endPos"] - value["startPos"])

                exec_statement = "INSERT INTO position_index VALUES (?,?,?,?,?)"
                c.execute(
                    exec_statement,
                    # add counter as a primary key
                    (counter, curr_zoom, curr_zoom, value["startPos"], value["endPos"]),
                )

                counter += 1
                break

            curr_zoom += 1

        curr_zoom = 0
    conn.commit()
    return True


###############################################################################


def _bedgraph(
    filepath,
    output_file,
    assembly,
    chrom_col,
    from_pos_col,
    to_pos_col,
    value_col,
    has_header,
    chromosome,
    tile_size,
    chunk_size,
    method,
    nan_value,
    transform,
    count_nan,
    closed_interval,
    chromsizes_filename,
    zoom_step,
):
    if output_file is None:
        output_file = op.splitext(filepath)[0] + ".hitile"

    print("output file:", output_file)

    # Override the output file if it existts
    if op.exists(output_file):
        os.remove(output_file)
    f = h5py.File(output_file, "w")

    # get the information about the chromosomes in this assembly
    if chromsizes_filename is not None:
        chrom_info = nc.get_chrominfo_from_file(chromsizes_filename)
        chrom_order = [
            a.encode("utf-8") for a in nc.get_chromorder_from_file(chromsizes_filename)
        ]
        chrom_sizes = nc.get_chromsizes_from_file(chromsizes_filename)
    else:
        chrom_info = nc.get_chrominfo(assembly)
        chrom_order = [a.encode("utf-8") for a in nc.get_chromorder(assembly)]
        chrom_sizes = nc.get_chromsizes(assembly)

    assembly_size = chrom_info.total_length
    print("assembly_size:", assembly_size)

    tile_size = tile_size
    # how many values to read in at once while tiling
    chunk_size = tile_size * 2 ** chunk_size

    dsets = []  # data sets at each zoom level
    nan_dsets = []  # store nan values

    # initialize the arrays which will store the values at each stored zoom
    # level
    z = 0
    positions = []  # store where we are at the current dataset
    data_buffers = [[]]
    nan_data_buffers = [[]]

    while assembly_size / 2 ** z > tile_size:
        dset_length = math.ceil(assembly_size / 2 ** z)
        dsets += [
            f.create_dataset(
                "values_" + str(z), (dset_length,), dtype="f", compression="gzip"
            )
        ]
        nan_dsets += [
            f.create_dataset(
                "nan_values_" + str(z), (dset_length,), dtype="f", compression="gzip"
            )
        ]

        data_buffers += [[]]
        nan_data_buffers += [[]]

        positions += [0]
        z += zoom_step

    # store some meta data
    d = f.create_dataset("meta", (1,), dtype="f")

    print("assembly:", assembly)

    d.attrs["zoom-step"] = zoom_step
    d.attrs["max-length"] = assembly_size
    d.attrs["assembly"] = assembly
    d.attrs["chrom-names"] = chrom_order
    d.attrs["chrom-sizes"] = chrom_sizes
    d.attrs["chrom-order"] = chrom_order
    d.attrs["tile-size"] = tile_size
    d.attrs["max-zoom"] = max_zoom = math.ceil(
        math.log(d.attrs["max-length"] / tile_size) / math.log(2)
    )
    d.attrs["max-width"] = tile_size * 2 ** max_zoom
    d.attrs["max-position"] = 0

    print("assembly size (max-length)", d.attrs["max-length"])
    print("max-width", d.attrs["max-width"])
    print("max_zoom:", d.attrs["max-zoom"])
    print("chunk-size:", chunk_size)
    print("chrom-order", d.attrs["chrom-order"])

    t1 = time.time()

    # are we reading the input from stdin or from a file?

    if filepath == "-":
        f = sys.stdin
    else:
        if filepath.endswith(".gz"):
            import gzip

            f = gzip.open(filepath, "rt")
        else:
            f = open(filepath, "r")

    curr_zoom = 0

    def add_values_to_data_buffers(buffers_to_add, nan_buffers_to_add):
        curr_zoom = 0

        data_buffers[0] += buffers_to_add
        nan_data_buffers[0] += nan_buffers_to_add

        curr_time = time.time() - t1
        percent_progress = (positions[curr_zoom] + 1) / float(assembly_size)
        print(
            "position: {} progress: {:.2f} elapsed: {:.2f} "
            "remaining: {:.2f}".format(
                positions[curr_zoom] + 1,
                percent_progress,
                curr_time,
                curr_time / (percent_progress) - curr_time,
            )
        )

        while len(data_buffers[curr_zoom]) >= chunk_size:
            # get the current chunk and store it, converting nans to 0
            print("len(data_buffers[curr_zoom])", len(data_buffers[curr_zoom]))
            curr_chunk = np.array(data_buffers[curr_zoom][:chunk_size])
            nan_curr_chunk = np.array(nan_data_buffers[curr_zoom][:chunk_size])

            """
            print("1cc:", sum(curr_chunk))
            print("1db:", data_buffers[curr_zoom][:chunk_size])
            print("1curr_chunk:", nan_curr_chunk)
            """
            print("positions[curr_zoom]:", positions[curr_zoom])

            curr_pos = positions[curr_zoom]
            dsets[curr_zoom][curr_pos : curr_pos + chunk_size] = curr_chunk
            nan_dsets[curr_zoom][curr_pos : curr_pos + chunk_size] = nan_curr_chunk

            # aggregate and store aggregated values in the next zoom_level's
            # data
            data_buffers[curr_zoom + 1] += list(
                ct.aggregate(curr_chunk, 2 ** zoom_step)
            )
            nan_data_buffers[curr_zoom + 1] += list(
                ct.aggregate(nan_curr_chunk, 2 ** zoom_step)
            )

            data_buffers[curr_zoom] = data_buffers[curr_zoom][chunk_size:]
            nan_data_buffers[curr_zoom] = nan_data_buffers[curr_zoom][chunk_size:]

            # data = data_buffers[curr_zoom+1]
            # nan_data = nan_data_buffers[curr_zoom+1]

            # do the same for the nan values buffers

            positions[curr_zoom] += chunk_size
            curr_zoom += 1

            if curr_zoom * zoom_step >= max_zoom:
                break

    values = []
    nan_values = []

    if has_header:
        f.readline()

    # the genome position up to which we've filled in values
    curr_genome_pos = 0

    # keep track of the previous value so that we can use it to fill in NAN
    # values
    # prev_value = 0

    for line in f:
        # each line should indicate a chromsome, start position and end
        # position
        parts = line.strip().split()

        start_genome_pos = chrom_info.cum_chrom_lengths[parts[chrom_col - 1]] + int(
            parts[from_pos_col - 1]
        )

        if start_genome_pos - curr_genome_pos > 1:
            values += [np.nan] * (start_genome_pos - curr_genome_pos - 1)
            nan_values += [1] * (start_genome_pos - curr_genome_pos - 1)

            curr_genome_pos += start_genome_pos - curr_genome_pos - 1

        # count how many nan values there are in the dataset
        nan_count = 1 if parts[value_col - 1] == nan_value else 0

        # if the provided values are log2 transformed, we have to un-transform
        # them
        if transform == "exp2":
            value = (
                2 ** float(parts[value_col - 1])
                if not parts[value_col - 1] == nan_value
                else np.nan
            )
        else:
            value = (
                float(parts[value_col - 1])
                if not parts[value_col - 1] == nan_value
                else np.nan
            )

        # we're going to add as many values are as specified in the bedfile line
        values_to_add = [value] * (
            int(parts[to_pos_col - 1]) - int(parts[from_pos_col - 1])
        )
        nan_counts_to_add = [nan_count] * (
            int(parts[to_pos_col - 1]) - int(parts[from_pos_col - 1])
        )

        if closed_interval:
            values_to_add += [value]
            nan_counts_to_add += [nan_count]

        # print("values_to_add", values_to_add)

        values += values_to_add
        nan_values += nan_counts_to_add

        d.attrs["max-position"] = start_genome_pos + len(values_to_add)

        curr_genome_pos += len(values_to_add)

        while len(values) > chunk_size:
            print("len(values):", len(values), chunk_size)
            print("line:", line)
            add_values_to_data_buffers(values[:chunk_size], nan_values[:chunk_size])
            values = values[chunk_size:]
            nan_values = nan_values[chunk_size:]

    add_values_to_data_buffers(values, nan_values)

    # store the remaining data
    while True:
        # get the current chunk and store it
        chunk_size = len(data_buffers[curr_zoom])
        curr_chunk = np.array(data_buffers[curr_zoom][:chunk_size])
        nan_curr_chunk = np.array(nan_data_buffers[curr_zoom][:chunk_size])

        """
        print("2curr_chunk", curr_chunk)
        print("2curr_zoom:", curr_zoom)
        print("2db", data_buffers[curr_zoom][:100])
        """

        curr_pos = positions[curr_zoom]
        dsets[curr_zoom][curr_pos : curr_pos + chunk_size] = curr_chunk
        nan_dsets[curr_zoom][curr_pos : curr_pos + chunk_size] = nan_curr_chunk

        # aggregate and store aggregated values in the next zoom_level's data
        data_buffers[curr_zoom + 1] += list(ct.aggregate(curr_chunk, 2 ** zoom_step))
        nan_data_buffers[curr_zoom + 1] += list(
            ct.aggregate(nan_curr_chunk, 2 ** zoom_step)
        )

        data_buffers[curr_zoom] = data_buffers[curr_zoom][chunk_size:]
        nan_data_buffers[curr_zoom] = nan_data_buffers[curr_zoom][chunk_size:]

        # data = data_buffers[curr_zoom+1]
        # nan_data = nan_data_buffers[curr_zoom+1]

        positions[curr_zoom] += chunk_size
        curr_zoom += 1

        # we've created enough tile levels to cover the entire maximum width
        if curr_zoom * zoom_step >= max_zoom:
            break

    # still need to take care of the last chunk


def _geojson(filepath, output_file, max_per_tile, tile_size, max_zoom):
    if filepath == "-":
        f = sys.stdin
    elif filepath.endswith(".gz"):
        f = gzip.open(filepath, "rt")
    else:
        f = open(filepath, "r")

    if output_file is None:
        output_file = filepath + ".gjdb"
    else:
        output_file = output_file

    if op.exists(output_file):
        os.remove(output_file)

    geojson = json.load(f)

    entries = []

    def getRect(coords, no_area_comp):
        minX = math.inf
        maxX = -math.inf
        minY = math.inf
        maxY = -math.inf
        area = 0.0  # Calculated with the shoelace formula
        n = len(coords)

        try:
            for coord_group in coords:
                for i, coord in enumerate(coord_group):
                    minX = min(minX, coord[0])
                    maxX = max(maxX, coord[0])
                    minY = min(minY, coord[1])
                    maxY = max(maxY, coord[1])
                    if not no_area_comp:
                        j = (i + 1) % n
                        area += coord_group[i][0] * coord_group[j][1]
                        area -= coord_group[j][0] * coord_group[i][1]

                area = abs(area) / 2.0
        except TypeError:
            # coords aren't iterable so this must be a point
            minX = coords[0]
            maxX = coords[0]
            minY = coords[1]
            maxY = coords[1]

            # points don't have an area so let's just pick something
            area = random.random()

        return minX, maxX, minY, maxY, abs(area) / 2.0

    for feature in geojson["features"]:
        try:
            area = feature["properties"]["area"]
        except:
            area = None
            pass

        try:
            uid = feature["properties"]["uid"]
        except:
            uid = None
            pass

        try:
            minLng, maxLng, minLat, maxLat, _area = getRect(
                feature["geometry"]["coordinates"], area
            )
            entries.append(
                {
                    "minLng": minLng,
                    "maxLng": maxLng,
                    "minLat": minLat,
                    "maxLat": maxLat,
                    "importance": area or _area,
                    "uid": uid or slugid.nice(),
                    "geometry": json.dumps(feature["geometry"]),
                    "properties": json.dumps(feature["properties"]),
                }
            )
        except Exception:
            raise

    # this script stores data in a sqlite database
    sqlite3.register_adapter(np.int64, lambda val: int(val))
    conn = sqlite3.connect(output_file)

    # store some meta data
    conn.execute(
        """
        CREATE TABLE tileset_info
        (
            zoom_step INT,
            tile_size INT,
            max_zoom INT,
            min_x INT,
            max_x INT,
            min_y INT,
            max_y INT
        )
        """
    )

    conn.execute(
        "INSERT INTO tileset_info VALUES (?,?,?,?,?,?,?)",
        (1, tile_size, 19, -180, 180, -90, 90),
    )
    conn.commit()

    # max_width = tile_size * 2 ** max_zoom
    # uid_to_entry = {}

    c = conn.cursor()
    c.execute(
        """
        CREATE TABLE intervals
        (
            id int PRIMARY KEY,
            zoomLevel int,
            importance real,
            minLng int,
            maxLng int,
            minLat int,
            maxLat int,
            uid text,
            geometry text,
            properties text
        )
        """
    )

    c.execute(
        """
        CREATE VIRTUAL TABLE position_index USING rtree(
            id,
            rMinLng, rMaxLng,
            rMinLat, rMaxLat
        )
        """
    )

    curr_zoom = 0
    counter = 0

    tile_counts = col.defaultdict(lambda: col.defaultdict(lambda: col.defaultdict(int)))
    entries = sorted(entries, key=lambda x: -x["importance"])

    counter = 0
    for d in entries:
        curr_zoom = 0

        while curr_zoom <= max_zoom:
            tile_from = get_tile_pos_from_lng_lat(d["minLng"], d["maxLat"], curr_zoom)

            tile_to = get_tile_pos_from_lng_lat(d["maxLng"], d["minLat"], curr_zoom)

            empty_tiles = True

            # go through and check if any of the tiles at this zoom level are
            # full
            for i in range(int(tile_from[0]), int(tile_to[0]) + 1):
                if not empty_tiles:
                    break

                for j in range(int(tile_from[1]), int(tile_to[1]) + 1):
                    if tile_counts[curr_zoom][i][j] > max_per_tile:

                        empty_tiles = False
                        break

            if empty_tiles:
                # they're all empty so add this interval to this zoom level
                for i in range(int(tile_from[0]), int(tile_to[0]) + 1):
                    for j in range(int(tile_from[1]), int(tile_to[1]) + 1):
                        tile_counts[curr_zoom][i][j] += 1

                c.execute(
                    "INSERT INTO intervals VALUES (?,?,?,?,?,?,?,?,?,?)",
                    (
                        counter,
                        curr_zoom,
                        d["importance"],
                        d["minLng"],
                        d["maxLng"],
                        d["minLat"],
                        d["maxLat"],
                        d["uid"],
                        d["geometry"],
                        d["properties"],
                    ),
                )
                conn.commit()

                c.execute(
                    "INSERT INTO position_index VALUES (?,?,?,?,?)",
                    (
                        # add counter as a primary key
                        counter,
                        d["minLng"],
                        d["maxLng"],
                        d["minLat"],
                        d["maxLat"],
                    ),
                )
                conn.commit()

                counter += 1
                break

            curr_zoom += 1

    f.close()


@aggregate.command()
@click.argument("filepath", metavar="FILEPATH")
@click.option(
    "--output-file",
    "-o",
    default=None,
    help="The default output file name to use. If this isn't "
    "specified, clodius will replace the current extension "
    "with .hitile",
)
@click.option(
    "--assembly",
    "-a",
    help="The genome assembly that this file was created against",
    type=click.Choice(nc.available_chromsizes()),
    default="hg19",
)
@click.option(
    "--chromosome",
    default=None,
    help="Only extract values for a particular chromosome. "
    "Use all chromosomes if not set.",
)
@click.option(
    "--tile-size",
    "-t",
    default=1024,
    help="The number of data points in each tile. "
    "Used to determine the number of zoom levels "
    "to create.",
)
@click.option(
    "--chunk-size",
    "-c",
    help="How many values to aggregate at once. "
    "Specified as a power of two multiplier of the tile "
    "size",
    default=14,
)
@click.option(
    "--chromosome-col",
    help="The column number (1-based) which contains the chromosome " "name",
    default=1,
)
@click.option(
    "--from-pos-col",
    help="The column number (1-based) which contains the starting " "position",
    default=2,
)
@click.option(
    "--to-pos-col",
    help="The column number (1-based) which contains the ending " "position",
    default=3,
)
@click.option(
    "--value-col",
    help="The column number (1-based) which contains the actual value " "position",
    default=4,
)
@click.option(
    "--has-header/--no-header",
    help="Does this file have a header that we should ignore",
    default=False,
)
@click.option(
    "--method",
    help="The method used to aggregate values (e.g. sum, average...)",
    type=click.Choice(["sum", "average"]),
    default="sum",
)
@click.option(
    "--nan-value", help="The string to use as a NaN value", type=str, default=None
)
@click.option(
    "--transform",
    help="The method used to aggregate values (e.g. sum, average...)",
    type=click.Choice(["none", "exp2"]),
    default="none",
)
@click.option(
    "--count-nan",
    help="Simply count the number of nan values in the file",
    is_flag=True,
)
@click.option(
    "--closed-interval", help="Treat the to column as a closed interval", is_flag=True
)
@click.option(
    "--chromsizes-filename",
    help="A file containing chromosome sizes and order",
    default=None,
)
@click.option(
    "--zoom-step",
    "-z",
    help="The number of intermediate aggregation levels to " "omit",
    default=8,
)
def bedgraph(
    filepath,
    output_file,
    assembly,
    chromosome_col,
    from_pos_col,
    to_pos_col,
    value_col,
    has_header,
    chromosome,
    tile_size,
    chunk_size,
    method,
    nan_value,
    transform,
    count_nan,
    closed_interval,
    chromsizes_filename,
    zoom_step,
):
    _bedgraph(
        filepath,
        output_file,
        assembly,
        chromosome_col,
        from_pos_col,
        to_pos_col,
        value_col,
        has_header,
        chromosome,
        tile_size,
        chunk_size,
        method,
        nan_value,
        transform,
        count_nan,
        closed_interval,
        chromsizes_filename,
        zoom_step,
    )


@aggregate.command()
@click.argument("filepath", metavar="FILEPATH")
@click.option(
    "--output-file",
    "-o",
    default=None,
    help="The default output file name to use. If this isn't "
    "specified, clodius will replace the current extension "
    "with .multires.bed",
)
@click.option(
    "--assembly",
    "-a",
    help="The genome assembly that this file was created against",
    default="hg19",
)
@click.option(
    "--importance-column",
    help="The column (1-based) containing information about how important "
    "that row is. If it's absent, then use the length of the region. "
    "If the value is equal to `random`, then a random value will be "
    "used for the importance (effectively leading to random sampling)",
)
@click.option(
    "--has-header/--no-header",
    help="Does this file have a header that we should ignore",
    default=False,
)
@click.option(
    "--chromosome",
    default=None,
    help="Only extract values for a particular chromosome. "
    "Use all chromosomes if not set.",
)
@click.option(
    "--max-per-tile",
    default=100,
    type=int,
    help="The maximum number of entries to store per tile",
)
@click.option(
    "--tile-size",
    default=1024,
    help="The number of nucleotides that the highest resolution tiles "
    "should span. This determines the maximum zoom level",
)
@click.option("--delimiter", default=None, type=str)
@click.option(
    "--chromsizes-filename",
    help="A file containing chromosome sizes and order",
    default=None,
)
@click.option(
    "--offset",
    help="Apply an offset to all the coordinates in this file",
    type=int,
    default=0,
)
def bedfile(
    filepath,
    output_file,
    assembly,
    importance_column,
    has_header,
    chromosome,
    max_per_tile,
    tile_size,
    delimiter,
    chromsizes_filename,
    offset,
):
    _bedfile(
        filepath,
        output_file,
        assembly,
        importance_column,
        has_header,
        chromosome,
        max_per_tile,
        tile_size,
        delimiter,
        chromsizes_filename,
        offset,
    )


@aggregate.command()
@click.argument("filepath", metavar="FILEPATH")
@click.option(
    "--output-file",
    "-o",
    default=None,
    help="The default output file name to use. If this isn't "
    "specified, clodius will replace the current extension "
    "with .bed2db",
)
@click.option(
    "--assembly",
    "-a",
    help="The genome assembly that this file was created against",
    default="hg19",
)
@click.option(
    "--importance-column",
    help="The column (1-based) containing information about how important "
    "that row is. If it's absent, then use the length of the region. "
    "If the value is equal to `random`, then a random value will be "
    "used for the importance (effectively leading to random sampling)",
    default="random",
)
@click.option(
    "--has-header/--no-header",
    help="Does this file have a header that we should ignore",
    default=False,
)
@click.option(
    "--max-per-tile",
    default=100,
    type=int,
    help="The maximum number of entries to include per tile",
)
@click.option(
    "--tile-size",
    default=1024,
    help="The number of nucleotides that the highest resolution tiles "
    "should span. This determines the maximum zoom level",
)
@click.option(
    "--chromosome",
    default=None,
    help="Only extract values for a particular chromosome. "
    "Use all chromosomes if not set.",
)
@click.option(
    "--chromsizes-filename",
    help="A file containing chromosome sizes and order",
    default=None,
)
@click.option(
    "--chr1-col", default=1, help="The column containing the first chromosome"
)
@click.option(
    "--chr2-col", default=4, help="The column containing the second chromosome"
)
@click.option(
    "--from1-col", default=2, help="The column containing the first start position"
)
@click.option(
    "--from2-col", default=5, help="The column containing the second start position"
)
@click.option(
    "--to1-col", default=3, help="The column containing the first end position"
)
@click.option(
    "--to2-col", default=6, help="The column containing the second end position"
)
def bedpe(
    filepath,
    output_file,
    assembly,
    importance_column,
    has_header,
    max_per_tile,
    tile_size,
    chromosome,
    chromsizes_filename,
    chr1_col,
    from1_col,
    to1_col,
    chr2_col,
    from2_col,
    to2_col,
):
    """Aggregate bedpe files"""
    _bedpe(
        filepath,
        output_file,
        assembly,
        importance_column,
        has_header,
        max_per_tile,
        tile_size,
        chromosome,
        chromsizes_filename,
        chr1_col=chr1_col,
        from1_col=from1_col,
        to1_col=to1_col,
        chr2_col=chr2_col,
        from2_col=from2_col,
        to2_col=to2_col,
    )


@aggregate.command()
@click.argument("filepath", metavar="FILEPATH")
@click.option(
    "-o",
    "--output-file",
    default=None,
    help="The default output file name to use. If this isn't "
    "specified, clodius will replace the current extension "
    "with .gjdb",
)
@click.option(
    "-a",
    "--assembly",
    default=None,
    help="The assembly that this data comes from. This parameter is "
    "unnecessary and/or overwritten if --chromsizes-filename is specified",
)
@click.option(
    "-s",
    "--tile-size",
    default=256,
    help="The number of nucleotides that the highest resolution tiles "
    "should span. This determines the maximum zoom level",
)
@click.option(
    "-c",
    "--chromsizes-filename",
    default=None,
    help="The file containnig chromosome sizes and order",
)
@click.option(
    "--starting-resolution",
    default=256,
    help="The resolution that the starting data is at (e.g. 1, 10, 20)",
)
@click.option(
    "--method",
    help="The method used to aggregate values (e.g. sum, average...)",
    type=click.Choice(["sum", "logsumexp"]),
    default="sum",
)
@click.option(
    "--row-infos-filename",
    help="A file containing the names of the rows in the multivec file",
    default=None,
)
def multivec(
    filepath,
    output_file,
    assembly,
    tile_size,
    chromsizes_filename,
    starting_resolution,
    method,
    row_infos_filename,
):
    """Aggregate a multivec file"""
    _multivec(
        filepath,
        output_file,
        assembly,
        tile_size,
        chromsizes_filename,
        starting_resolution,
        method,
        row_infos_filename,
    )


@aggregate.command()
@click.argument("filepath", metavar="FILEPATH")
@click.option(
    "-o",
    "--output-file",
    default=None,
    help="The default output file name to use. If this isn't "
    "specified, clodius will replace the current extension "
    "with .gjdb",
)
@click.option("-m", "--max-per-tile", default=20, type=int)
@click.option(
    "-s",
    "--tile-size",
    default=256,
    help="The number of nucleotides that the highest resolution tiles "
    "should span. This determines the maximum zoom level",
)
@click.option(
    "-z",
    "--max-zoom",
    default=19,
    help="The number of nucleotides that the highest resolution tiles "
    "should span. This determines the maximum zoom level",
)
def geojson(filepath, output_file, max_per_tile, tile_size, max_zoom):
    """Aggregate a geojson file"""
    _geojson(filepath, output_file, max_per_tile, tile_size, max_zoom)



import os.path as op


def partition_by_adjacent_tiles(tile_ids, dimension=2):
    """
    Partition a set of tile ids into sets of adjacent tiles. For example,
    if we're requesting a set of four tiles that form a rectangle, then
    those four tiles will become one set of adjacent tiles. Non-contiguous
    tiles are not grouped together.

    Parameters
    ----------
    tile_ids: [str,...]
        A list of tile_ids (e.g. xyx.0.0.1) identifying the tiles
        to be retrieved
    dimension: int
        The dimensionality of the tiles

    Returns
    -------
    tile_lists: [tile_ids, tile_ids]
        A list of tile lists, all of which have tiles that
        are within 1 position of another tile in the list
    """
    tile_id_lists = []

    for tile_id in sorted(
        tile_ids, key=lambda x: [int(p) for p in x.split(".")[2 : 2 + dimension]]
    ):
        tile_id_parts = tile_id.split(".")

        # exclude the zoom level in the position
        # because the tiles should already have been partitioned
        # by zoom level
        tile_position = list(map(int, tile_id_parts[2:4]))

        added = False

        for tile_id_list in tile_id_lists:
            # iterate over each group of adjacent tiles
            for ct_tile_id in tile_id_list:
                ct_tile_id_parts = ct_tile_id.split(".")
                ct_tile_position = list(map(int, ct_tile_id_parts[2 : 2 + dimension]))
                far_apart = False

                # iterate over each dimension and see if this tile is close
                for p1, p2 in zip(tile_position, ct_tile_position):
                    if abs(int(p1) - int(p2)) > 1:
                        # too far apart can't be part of the same group
                        far_apart = True

                if not far_apart:
                    # no position was too far
                    tile_id_list += [tile_id]
                    added = True
                    break

            if added:
                break
        if not added:
            tile_id_lists += [[tile_id]]

    return tile_id_lists


def infer_filetype(filename):
    _, ext = op.splitext(filename)

    if ext.lower() == ".bw" or ext.lower() == ".bigwig":
        return "bigwig"
    elif ext.lower() == ".mcool" or ext.lower() == ".cool":
        return "cooler"
    elif ext.lower() == ".htime":
        return "time-interval-json"
    elif ext.lower() == ".hitile":
        return "hitile"
    elif ext.lower() == ".beddb":
        return "beddb"
    elif ext.lower() == ".mv5":
        return "multivec"

    return None


def infer_datatype(filetype):
    if filetype == "cooler":
        return "matrix"
    if filetype == "bigwig":
        return "vector"
    if filetype == "time-interval-json":
        return "time-interval"
    if filetype == "hitile":
        return "vector"
    if filetype == "beddb":
        return "bedlike"


def tiles_wrapper_2d(tile_ids, tiles_function):
    tile_values = []

    for tile_id in tile_ids:
        parts = tile_id.split(".")

        if len(parts) < 4:
            raise IndexError("Not enough tile info present")

        z, x, y = map(int, [parts[1], parts[2], parts[3]])

        tile_values += [(tile_id, tiles_function(z, x, y))]

    return tile_values


def bundled_tiles_wrapper_2d(tile_ids, tiles_function):
    """
    Bundle adjacent tile requests so that they can be
    processed concurrently. This is helpful for function
    that require scanning a dataset. It's faster to filter
    a large region and then break it down into individual
    tiles than to go over the entire dataset and filter
    individual tiles multiple times.
    """
    tile_values = []

    partitioned_tile_lists = partition_by_adjacent_tiles(tile_ids)

    for tile_group in partitioned_tile_lists:
        zoom_level = int(tile_group[0].split(".")[1])
        tileset_id = tile_group[0].split(".")[0]

        tile_positions = [[int(x) for x in t.split(".")[2:4]] for t in tile_group]

        minx = min([t[0] for t in tile_positions])
        maxx = max([t[0] for t in tile_positions])

        miny = min([t[1] for t in tile_positions])
        maxy = max([t[1] for t in tile_positions])

        tf = tiles_function(
            zoom_level, minx, miny, width=maxx - minx + 1, height=maxy - miny + 1
        )

        tile_values += [
            ("{}.{}".format(tileset_id, ".".join(map(str, tile_position))), data)
            for (tile_position, data) in tf
        ]

    return tile_values


def tile_bounds(tsinfo, z, x, y, width=1, height=1):
    """
    Get the coordinate boundaries for the given tile.

    Parameters:
    -----------
    tsinfo: { min_pos: [], max_pos [] }
        Tileset info containing the bounds of the dataset
    z: int
        The zoom level
    x: int
        The x position
    y: int
        The y position
    width: int
        Return bounds for a region encompassing multiple tiles
    height: int
        Return bounds for a region encompassing multiple tiles
    """
    min_pos = tsinfo["min_pos"]
    max_pos = tsinfo["max_pos"]

    max_width = max(max_pos[0] - min_pos[0], max_pos[1] - min_pos[1])

    tile_width = max_width / 2 ** z
    from_x = min_pos[0] + x * tile_width
    to_x = min_pos[0] + (x + width) * tile_width

    from_y = min_pos[1] + y * tile_width
    to_y = min_pos[1] + (y + height) * tile_width

    return [from_x, from_y, to_x, to_y]



import h5py
import logging

logger = logging.getLogger(__name__)


def get_cooler_info(file_path):
    """Get information of a cooler file.

    Args:
        file_path (str): Path to a cooler file.

    Returns:
        dict: Dictionary containing basic information about the cooler file.
    """
    TILE_SIZE = 256
    CHROM_CUM_LEN = 0
    with h5py.File(file_path, "r") as f:
        max_zoom = f.attrs.get("max-zoom")

        if max_zoom is None:
            logger.info("no zoom found")
            raise ValueError("The `max_zoom` attribute is missing.")

        total_length = CHROM_CUM_LEN
        max_zoom = f.attrs["max-zoom"]
        bin_size = int(f[str(max_zoom)].attrs["bin-size"])

        max_width = bin_size * TILE_SIZE * 2 ** max_zoom

        info = {
            "min_pos": [0.0, 0.0],
            "max_pos": [total_length, total_length],
            "max_zoom": max_zoom,
            "max_width": max_width,
            "bins_per_dimension": TILE_SIZE,
        }

        return info



import csv
import logging

logger = logging.getLogger(__name__)


def get_tsv_chromsizes(filename):
    """
    Get a list of chromosome sizes from this [presumably] tsv
    chromsizes file file.

    Parameters:
    -----------
    filename: string
        The filename of the tsv file

    Returns
    -------
    chromsizes: [(name:string, size:int), ...]
        An ordered list of chromosome names and sizes
    """
    try:
        with open(filename, "r") as f:
            reader = csv.reader(f, delimiter="\t")

            data = []
            for row in reader:
                data.append(row)
        return data
    except Exception as ex:
        logger.error(ex)

        err_msg = "WHAT?! Could not load file %s. 😤 (%s)" % (filename, ex)

        raise Exception(err_msg)



import base64
import numpy as np


def format_dense_tile(data):
    """
    Format raw vector data into a more structured tile
    containing either float16 or float32 data along with a
    dtype to differentiate between the two.

    Parameters
    ----------
    tile_data_array: np.array
        An array of values
    Returns
    -------
    tile_data: {'dense': str, 'dtype': str}
        The tile data reformatted to use float16 or float32 as the
        datatype. The dtype indicates which format is chosen.
    """

    tile_data = {}

    if len(data):
        with np.warnings.catch_warnings():
            np.warnings.filterwarnings("ignore", r"All-NaN (slice|axis) encountered")

            max_dense = float(np.nanmax(data))
            min_dense = float(np.nanmin(data))
    else:
        max_dense = np.nan
        min_dense = np.nan

    tile_data["min_value"] = min_dense if not np.isnan(min_dense) else "NaN"
    tile_data["max_value"] = max_dense if not np.isnan(max_dense) else "NaN"

    min_f16 = np.finfo("float16").min
    max_f16 = np.finfo("float16").max

    has_nan = np.sum(np.isnan(data)) > 0
    n_dim = len(data.shape)
    size = data.shape[1] if n_dim == 2 else 1

    # Flatten data array
    fdata = data
    if n_dim > 0:
        fdata = data.flatten()

    if (
        not has_nan
        and max_dense > min_f16
        and max_dense < max_f16
        and min_dense > min_f16
        and min_dense < max_f16
    ):
        tile_data.update(
            {
                "dense": base64.b64encode(fdata.astype("float16")).decode("utf-8"),
                "size": size,
                "dtype": "float16",
            }
        )
    else:
        tile_data.update(
            {
                "dense": base64.b64encode(fdata.astype("float32")).decode("utf-8"),
                "size": size,
                "dtype": "float32",
            }
        )

    return tile_data



import clodius.tiles.cooler as hgco
import numpy as np
import os.path as op
import base64


def test_cooler_info():
    filename = op.join("data", "Dixon2012-J1-NcoI-R1-filtered.100kb.multires.cool")

    hgco.tileset_info(filename)
    # TODO: Make assertion about info returned.

    tiles = hgco.generate_tiles(filename, ["a.0.0.0"])

    r = base64.decodestring(tiles[0][1]["dense"].encode("utf-8"))
    q = np.frombuffer(r, dtype=np.float32)

    q = q.reshape((256, 256))

    filename = op.join("data", "hic-resolutions.cool")
    # print(hgco.tileset_info(filename))


def test_cooler_tiles():
    filename = op.join("data", "hic-resolutions.cool")
    hgco.tiles(filename, ["x.0.0.0"])



import math
import numpy as np
import clodius.tiles.format as hgfo


def tiles_wrapper(grid, tile_ids):
    tile_values = []

    for tile_id in tile_ids:
        parts = tile_id.split(".")

        if len(parts) < 4:
            raise IndexError("Not enough tile info present")

        z = int(parts[1])
        x = int(parts[3])
        y = int(parts[2])

        ret_array = tiles(grid, z, x, y).reshape((-1))

        tile_values += [(tile_id, hgfo.format_dense_tile(ret_array))]

    return tile_values


def tileset_info(grid, bounds=None):
    """
    Get the tileset info for the grid
    """
    bin_size = 256
    max_dim = max(grid.shape)
    # print("grid.shape:", grid.shape)
    max_zoom = math.ceil(math.log(max_dim / bin_size) / math.log(2))
    max_zoom = 0 if max_zoom < 0 else max_zoom

    max_width = 2 ** max_zoom * bin_size
    max_width1 = 2 ** max_zoom * bin_size

    scale_up = max_width / max_dim

    if bounds is not None:
        min_pos = [bounds[0], bounds[1]]
        max_pos = [bounds[2], bounds[3]]

        # print('scale_up:', scale_up, max_pos[0] - min_pos[0])

        max_width = (max_pos[0] - min_pos[0]) * scale_up
        max_width1 = (max_pos[1] - min_pos[1]) * scale_up
    else:
        min_pos = [0, 0]
        max_pos = grid.shape

    if len(grid.shape) > 2:
        raise ValueError("Grid's shape is not conducive to plotting", grid.shape)
    return {
        "max_width": max_width,
        "max_width1": max_width1,
        "min_pos": min_pos,
        "max_pos": max_pos,
        "max_zoom": max_zoom,
        "mirror_tiles": "false",
        "bins_per_dimension": bin_size,
    }


def tiles(grid, z, x, y, nan_grid=None, bin_size=256):
    """
    Return tiles at the given positions.

    Parameters
    -----------
    grid: np.array
        An nxn array containing values
    z: int
        The zoom level (0 corresponds to most zoomed out)
    x: int
        The x tile position
    y: int
        The y tile position
    bin_size: int
        The number of values per bin
    """
    max_dim = max(grid.shape)
    # print("max_dim", max_dim)

    max_zoom = math.ceil(math.log(max_dim / bin_size) / math.log(2))
    max_zoom = 0 if max_zoom < 0 else max_zoom

    # max_width = 2 ** max_zoom * bin_size
    # print("max_width:", max_width, 'bin_size:', bin_size, 'max_zoom', max_zoom)

    tile_width = 2 ** (max_zoom - z) * bin_size

    x_start = x * tile_width
    y_start = y * tile_width

    x_end = min(grid.shape[0], x_start + tile_width)
    y_end = min(grid.shape[1], y_start + tile_width)

    # print("tile_width", tile_width)
    # print("x_start:", x_start, x_end)
    # print("y_start:", y_start, y_end)

    num_to_sum = 2 ** (max_zoom - z)
    # print("num_to_sum", num_to_sum)

    data = grid[x_start:x_end, y_start:y_end]
    # print("data:", data)

    # add some data so that the data can be divided into squares
    divisible_x_width = num_to_sum * math.ceil(data.shape[0] / num_to_sum)
    divisible_y_width = num_to_sum * math.ceil(data.shape[1] / num_to_sum)

    divisible_x_pad = divisible_x_width - data.shape[0]
    divisible_y_pad = divisible_y_width - data.shape[1]
    # print("data.shape", data.shape)

    # print("divisible_x_pad:", divisible_x_pad)
    # print("divisible_y_pad:", divisible_y_pad)

    a = np.pad(
        data,
        ((0, divisible_x_pad), (0, divisible_y_pad)),
        "constant",
        constant_values=(np.nan, np.nan),
    )

    b = np.nansum(a.reshape((a.shape[0], -1, num_to_sum)), axis=2)
    ret_array = np.nansum(b.T.reshape(b.shape[1], -1, num_to_sum), axis=2).T
    ret_array[ret_array == 0.0] = np.nan
    # print('ret_array:', ret_array)

    # print("sum:", np.nansum(ret_array))

    if nan_grid is not None:
        # print("normalizing")
        # we want to calculate the means of the data points

        # NOTE: In the line below, "nan_grid" was originally "not_nan_grid",
        # which is undefined. This is my best guess of the desired behavior.
        not_nan_data = nan_grid[x_start:x_end, y_start:y_end]
        na = np.pad(
            not_nan_data,
            ((0, divisible_x_pad), (0, divisible_y_pad)),
            "constant",
            constant_values=(np.nan, np.nan),
        )
        nb = np.nansum(na.reshape((na.shape[1], -1, num_to_sum)), axis=2)
        norm_array = np.nansum(nb.T.reshape(nb.shape[1], -1, num_to_sum), axis=2).T

        ret_array = ret_array / norm_array

    # determine how much to pad the array
    x_pad = bin_size - ret_array.shape[0]
    y_pad = bin_size - ret_array.shape[1]

    # print("ret_array:", ret_array.shape)
    # print("x_pad:", x_pad, "y_pad:", y_pad)

    return np.pad(
        ret_array,
        ((0, x_pad), (0, y_pad)),
        "constant",
        constant_values=(np.nan, np.nan),
    )



import bbi
import functools as ft
import logging
import numpy as np
import pandas as pd
import random
import clodius.tiles.bigwig as hgbw

from concurrent.futures import ThreadPoolExecutor

DEFAULT_RANGE_MODE = "significant"
MIN_ELEMENTS = 1
MAX_ELEMENTS = 200
DEFAULT_SCORE = 0

logger = logging.getLogger(__name__)

range_modes = {}
range_modes["significant"] = {"name": "Significant", "value": "significant"}


def tileset_info(bbpath, chromsizes=None):
    ti = hgbw.tileset_info(bbpath, chromsizes)
    ti["range_modes"] = range_modes
    return ti


def fetch_data(a):
    (
        bbpath,
        binsize,
        chromsizes,
        range_mode,
        min_elements,
        max_elements,
        cid,
        start,
        end,
    ) = a

    """
    Retrieve tile data from a bigbed file.

    This approach currently returns a subset of intervals within the bounds of the specified
    query range.

    The subset is determined, at this time, by using the population of scores in the score
    column of the BED data to generate a quantile value that would allow, at most, a maximum
    number of elements (either a default or specified value). Because intervals are discrete
    elements, it is possible for a quantile to allow a few more elements than the desired
    limit; in this case, a uniformly-random sample is drawn from the thresholded set without
    replacement.

    Parameters
    ----------
    bbpath: string
        The path to the bigBed media file
    binsize: integer
        Resolution of a bin at a particular zoom level
    chromsizes: [[chrom, size],...]
        A 2d array containing chromosome names and sizes. Overrides the
        chromsizes in chromsizes_map
    range_mode: string or None
        If specified, determines what rule is applied to intervals retrieved
        over the specified chromosome, start, and end range
    min_elements: integer
        For fetched intervals, return no fewer than the specified number
    max_elements: integer
        For fetched intervals, return no more than the specified number
    cid: integer
        Index of chromosome associated with chromsizes
    start: integer
        Start position of interval query (relative to chromosome)
    end: integer
        End position of interval query (relative to chromosome)

    Returns
    -------
    intervals: [{'chrOffset': integer, 'importance': integer, 'fields': [interval]}, ... ]
        A list of beddb-like gene annotation objects
    """

    try:
        chrom = chromsizes.index[cid]

        fetch_factory = ft.partial(bbi.fetch_intervals, bbpath, chrom, start, end)

        if range_mode == "significant":
            intervals, intervals2 = fetch_factory(), fetch_factory()
        else:
            intervals, intervals2 = fetch_factory(), fetch_factory()

    except IndexError:
        # beyond the range of the available chromosomes
        # probably means we've requested a range of absolute
        # coordinates that stretch beyond the end of the genome
        intervals, intervals2 = None, None

    except KeyError:
        # probably requested a chromosome that doesn't exist (e.g. chrM)
        intervals, intervals2 = None, None

    offset = 0
    offsetIdx = 0
    chrOffsets = {}
    for chrSize in chromsizes:
        chrOffsets[chromsizes.index[offsetIdx]] = offset
        offset += chrSize
        offsetIdx += 1

    final_intervals = []
    intervals_length = 0
    scores = []

    if not intervals:
        return final_intervals

    for interval in intervals:
        try:
            scores.append(int(interval[4]))
        except (ValueError, IndexError):
            scores.append(DEFAULT_SCORE)
        intervals_length += 1

    # generate beddb-like elements for parsing by the higlass plugin
    if intervals_length >= min_elements and intervals_length <= max_elements:
        for interval in intervals2:
            try:
                score = int(interval[4])
                final_intervals.append(
                    {
                        "chrOffset": chrOffsets[chrom],
                        "importance": score,
                        "fields": interval,
                    }
                )
            except (ValueError, IndexError):
                final_intervals.append(
                    {
                        "chrOffset": chrOffsets[chrom],
                        "importance": DEFAULT_SCORE,
                        "fields": interval,
                    }
                )

    elif intervals_length > max_elements:
        thresholded_intervals = []
        desired_perc = max_elements / intervals_length
        thresholded_score = int(np.quantile(scores, 1 - desired_perc))
        for interval in intervals2:
            try:
                score = int(interval[4])
                if score >= thresholded_score:
                    thresholded_intervals.append(
                        {
                            "chrOffset": chrOffsets[chrom],
                            "importance": score,
                            "fields": interval,
                        }
                    )
            except (ValueError, IndexError):
                if DEFAULT_SCORE >= thresholded_score:
                    thresholded_intervals.append(
                        {
                            "chrOffset": chrOffsets[chrom],
                            "importance": DEFAULT_SCORE,
                            "fields": interval,
                        }
                    )
        thresholded_intervals_length = len(thresholded_intervals)
        if thresholded_intervals_length > max_elements:
            indices = random.sample(range(thresholded_intervals_length), max_elements)
            final_intervals = [thresholded_intervals[i] for i in sorted(indices)]

    return final_intervals


def get_bigbed_tile(
    bbpath,
    zoom_level,
    start_pos,
    end_pos,
    chromsizes=None,
    range_mode=None,
    min_elements=None,
    max_elements=None,
):
    if chromsizes is None:
        chromsizes = hgbw.get_chromsizes(bbpath)

    if min_elements is None:
        min_elements = MIN_ELEMENTS
    if max_elements is None:
        max_elements = MAX_ELEMENTS

    resolutions = hgbw.get_zoom_resolutions(chromsizes)
    binsize = resolutions[zoom_level]

    cids_starts_ends = list(hgbw.abs2genomic(chromsizes, start_pos, end_pos))

    with ThreadPoolExecutor(max_workers=16) as e:
        arrays = list(
            e.map(
                fetch_data,
                [
                    tuple(
                        [
                            bbpath,
                            binsize,
                            chromsizes,
                            range_mode,
                            min_elements,
                            max_elements,
                        ]
                        + list(c)
                    )
                    for c in cids_starts_ends
                ],
            )
        )

    # concatenate bigBed tileset data across chromosomes, so that it looks similar to a beddb response
    results = [x for x in arrays if x != []]
    return [item for sublist in results for item in sublist]


def tiles(bbpath, tile_ids, chromsizes_map={}, chromsizes=None):
    """
    Generate tiles from a bigbed file.

    Parameters
    ----------
    tileset: tilesets.models.Tileset object
        The tileset that the tile ids should be retrieved from
    tile_ids: [str,...]
        A list of tile_ids (e.g. xyx.0.0) identifying the tiles
        to be retrieved
    chromsizes_map: {uid: []}
        A set of chromsizes listings corresponding to the parameters of the
        tile_ids. To be used if a chromsizes id is passed in with the tile id
        with the `|cos:id` tag in the tile id
    chromsizes: [[chrom, size],...]
        A 2d array containing chromosome names and sizes. Overrides the
        chromsizes in chromsizes_map

    Returns
    -------
    tile_list: [(tile_id, tile_data),...]
        A list of tile_id, tile_data tuples
    """

    min_elements = -1
    max_elements = -1

    generated_tiles = []
    for tile_id in tile_ids:
        tile_option_parts = tile_id.split("|")[1:]
        tile_no_options = tile_id.split("|")[0]
        tile_id_parts = tile_no_options.split(".")
        tile_position = list(map(int, tile_id_parts[1:3]))
        return_value = (
            tile_id_parts[3] if len(tile_id_parts) > 3 else DEFAULT_RANGE_MODE
        )

        range_mode = return_value if return_value in range_modes else None

        tile_options = dict([o.split(":") for o in tile_option_parts])

        if "min" in tile_options:
            min_elements = int(tile_options["min"])
        if "max" in tile_options:
            max_elements = int(tile_options["max"])

        if min_elements > max_elements:
            temp_min_elements = min_elements
            min_elements = max_elements
            max_elements = temp_min_elements
        elif min_elements == max_elements:
            min_elements = max_elements
            max_elements = min_elements + 1

        if max_elements <= 0:
            min_elements = MIN_ELEMENTS
            max_elements = MAX_ELEMENTS

        if chromsizes:
            chromnames = [c[0] for c in chromsizes]
            chromlengths = [int(c[1]) for c in chromsizes]
            chromsizes_to_use = pd.Series(chromlengths, index=chromnames)
        else:
            chromsizes_id = None
            if "cos" in tile_options:
                chromsizes_id = tile_options["cos"]
            if chromsizes_id in chromsizes_map:
                chromsizes_to_use = chromsizes_map[chromsizes_id]
            else:
                chromsizes_to_use = None

        zoom_level = tile_position[0]
        tile_pos = tile_position[1]

        # this doesn't combine multiple consequetive ids, which
        # would speed things up
        if chromsizes_to_use is None:
            chromsizes_to_use = hgbw.get_chromsizes(bbpath)

        max_depth = hgbw.get_quadtree_depth(chromsizes_to_use)
        tile_size = hgbw.TILE_SIZE * 2 ** (max_depth - zoom_level)
        start_pos = tile_pos * tile_size
        end_pos = start_pos + tile_size

        tile_value = get_bigbed_tile(
            bbpath,
            zoom_level,
            start_pos,
            end_pos,
            chromsizes_to_use,
            range_mode=range_mode,
            min_elements=min_elements,
            max_elements=max_elements,
        )

        generated_tiles += [(tile_id, tile_value)]

    return generated_tiles


def chromsizes(filename):
    return hgbw.chromsizes(filename)



from __future__ import print_function

import clodius.db_tiles as cdt


def test_get_tileset_info():
    filename = "test/sample_data/arrowhead_domains_short.txt.multires.db"
    t = cdt.get_tileset_info(filename)

    assert t["zoom_step"] == 1
    assert t["max_length"] == 3095693981
    assert t["max_width"] > 4000000000
    assert t["max_width"] < 5000000000


def test_get_tiles():
    filename = "test/sample_data/arrowhead_domains_short.txt.multires.db"

    cdt.get_2d_tiles(filename, 0, 0, 0, numx=1, numy=1)
    # TODO: Make an assertion



import os
import shortuuid
import time

import_time = time.time()


def describe_dataset(args_raw, args_parsed):
    """
    Create a JSON structure describing how this dataset
    was created. The input is the parsed set of command line arguments used to
    run this program.

    :param args: The arguments used to run the program.
    :return: A JSON file with the arguments as well as some other identifying information.
    """
    return {
        "uuid": shortuuid.uuid(),
        "cwd": os.getcwd(),
        "args_raw": args_raw,
        "command": " ".join(args_raw),
        "args_parsed": vars(args_parsed),
        "import_time": time.strftime(
            "%Y-%m-%d %H:%M:%S %z", time.localtime(import_time)
        ),
        "current_time": time.strftime("%Y-%m-%d %H:%M:%S %z"),
        "elapsed_time": round(time.time() - import_time, 2),
    }



import clodius.tiles.bigwig as hgbi
import os.path as op
import numpy as np
import base64


def test_bigwig_tiles():
    filename = op.join(
        "data", "wgEncodeCaltechRnaSeqHuvecR1x75dTh1014IlnaPlusSignalRep2.bigWig"
    )

    mean_tile = hgbi.tiles(filename, ["x.0.0"])
    mean_mean_tile = hgbi.tiles(filename, ["x.0.0.mean"])
    min_tile = hgbi.tiles(filename, ["x.0.0.min"])
    max_tile = hgbi.tiles(filename, ["x.0.0.max"])
    std_tile = hgbi.tiles(filename, ["x.0.0.std"])

    assert mean_tile[0][1]["max_value"] == mean_mean_tile[0][1]["max_value"]
    assert mean_tile[0][1]["max_value"] > min_tile[0][1]["max_value"]
    assert max_tile[0][1]["max_value"] > mean_tile[0][1]["max_value"]
    assert (
        max_tile[0][1]["max_value"]
        > mean_tile[0][1]["max_value"] + std_tile[0][1]["max_value"]
    )

    min_max_tile = hgbi.tiles(filename, ["x.0.0.minMax"])
    whisker_tile = hgbi.tiles(filename, ["x.0.0.whisker"])

    mean_val = np.frombuffer(
        base64.b64decode(mean_tile[0][1]["dense"]), dtype=mean_tile[0][1]["dtype"]
    )

    min_val = np.frombuffer(
        base64.b64decode(min_tile[0][1]["dense"]), dtype=min_tile[0][1]["dtype"]
    )

    max_val = np.frombuffer(
        base64.b64decode(max_tile[0][1]["dense"]), dtype=max_tile[0][1]["dtype"]
    )

    std_val = np.frombuffer(
        base64.b64decode(std_tile[0][1]["dense"]), dtype=std_tile[0][1]["dtype"]
    )

    min_max_val = np.frombuffer(
        base64.b64decode(min_max_tile[0][1]["dense"]), dtype=min_max_tile[0][1]["dtype"]
    )

    mean_tile = hgbi.tiles(filename, ["x.0.0"])
    mean_mean_tile = hgbi.tiles(filename, ["x.0.0.mean"])
    min_tile = hgbi.tiles(filename, ["x.0.0.min"])
    max_tile = hgbi.tiles(filename, ["x.0.0.max"])
    std_tile = hgbi.tiles(filename, ["x.0.0.std"])

    assert mean_tile[0][1]["max_value"] == mean_mean_tile[0][1]["max_value"]
    assert mean_tile[0][1]["max_value"] > min_tile[0][1]["max_value"]
    assert max_tile[0][1]["max_value"] > mean_tile[0][1]["max_value"]
    assert (
        max_tile[0][1]["max_value"]
        > mean_tile[0][1]["max_value"] + std_tile[0][1]["max_value"]
    )

    min_max_tile = hgbi.tiles(filename, ["x.0.0.minMax"])
    whisker_tile = hgbi.tiles(filename, ["x.0.0.whisker"])

    mean_val = np.frombuffer(
        base64.b64decode(mean_tile[0][1]["dense"]), dtype=mean_tile[0][1]["dtype"]
    )

    min_val = np.frombuffer(
        base64.b64decode(min_tile[0][1]["dense"]), dtype=min_tile[0][1]["dtype"]
    )

    max_val = np.frombuffer(
        base64.b64decode(max_tile[0][1]["dense"]), dtype=max_tile[0][1]["dtype"]
    )

    std_val = np.frombuffer(
        base64.b64decode(std_tile[0][1]["dense"]), dtype=std_tile[0][1]["dtype"]
    )

    min_max_val = np.frombuffer(
        base64.b64decode(min_max_tile[0][1]["dense"]), dtype=min_max_tile[0][1]["dtype"]
    )

    whisker_val = np.frombuffer(
        base64.b64decode(whisker_tile[0][1]["dense"]), dtype=whisker_tile[0][1]["dtype"]
    )

    assert min_max_val.shape[0] == 2 * mean_val.shape[0]
    assert np.allclose(min_max_val[::2], min_val, equal_nan=True)
    assert np.allclose(min_max_val[1::2], max_val, equal_nan=True)

    assert whisker_val.shape[0] == 4 * mean_val.shape[0]
    assert np.allclose(whisker_val[::4], min_val, equal_nan=True)
    assert np.allclose(whisker_val[1::4], max_val, equal_nan=True)
    assert np.allclose(whisker_val[2::4], mean_val, equal_nan=True)
    assert np.allclose(whisker_val[3::4], std_val, equal_nan=True)


def test_tileset_info():
    filename = op.join(
        "data", "wgEncodeCaltechRnaSeqHuvecR1x75dTh1014IlnaPlusSignalRep2.bigWig"
    )

    tileset_info = hgbi.tileset_info(filename)

    assert len(tileset_info["aggregation_modes"]) == 4
    assert tileset_info["aggregation_modes"]["mean"]
    assert tileset_info["aggregation_modes"]["min"]
    assert tileset_info["aggregation_modes"]["max"]
    assert tileset_info["aggregation_modes"]["std"]

    assert len(tileset_info["range_modes"]) == 2
    assert tileset_info["range_modes"]["minMax"]
    assert tileset_info["range_modes"]["whisker"]


def test_natsorted():
    chromname_tests = [
        {
            "input": ["2", "3", "4", "m", "x", "1", "y"],
            "expected": ["1", "2", "3", "4", "x", "y", "m"],
        },
        {
            "input": ["chr1", "chr4", "chr5", "chr2", "chr3", "chrMT", "chrY", "chrX"],
            "expected": [
                "chr1",
                "chr2",
                "chr3",
                "chr4",
                "chr5",
                "chrX",
                "chrY",
                "chrMT",
            ],
        },
    ]

    for test in chromname_tests:
        sorted_output = hgbi.natsorted(test["input"])
        assert (
            sorted_output == test["expected"]
        ), "Sorted output was %s\nExpected: %s" % (sorted_output, test["expected"])



import clodius.tiles.files as hgfi


def test_infer_filetype():
    assert hgfi.infer_filetype("matrix.mcool") == "cooler"
    assert hgfi.infer_filetype("matrix.txt") is None
    assert hgfi.infer_filetype("a.bigwig") == "bigwig"
    assert hgfi.infer_filetype("a.bigWig") == "bigwig"


def test_infer_datatype():
    assert hgfi.infer_datatype("cooler") == "matrix"
    assert hgfi.infer_datatype("bigwig") == "vector"



from __future__ import print_function

import click.testing as clt
import clodius.cli.aggregate as cca
import clodius.tiles.beddb as ctb
import os
import os.path as op
import sqlite3
import tempfile

testdir = op.realpath(op.dirname(__file__))


def test_nonstandard_chrom():
    filename = "test/sample_data/test_non_standard_chrom.bed"
    f = tempfile.NamedTemporaryFile(delete=False)

    ret = cca._bedfile(
        filename, f.name, "hg19", None, False, None, 100, 1024, None, None, 0
    )

    assert ret is None


def test_get_tileset_info():
    filename = "test/sample_data/gene_annotations.short.db"
    t = ctb.tileset_info(filename)

    assert t["zoom_step"] == 1
    assert t["max_length"] == 3137161264
    assert t["max_width"] > 4000000000
    assert t["max_width"] < 5000000000


def test_table_created():
    check_table("test/sample_data/gene_annotations.short.db")


def check_table(filename):
    conn = sqlite3.connect(filename)
    c = conn.cursor()

    rows = c.execute(
        "SELECT * from intervals,position_index "
        "where intervals.id=position_index.id "
        "and zoomLevel < 1 "
        "and rStartPos > 2400000000 "
        "and rEndPos < 2500000000"
    )
    counter = 0
    for row in rows:
        assert row[3] > 2400000000
        assert row[4] < 2500000000
        counter += 1

    assert counter > 0


def test_get_tiles():
    filename = "test/sample_data/gene_annotations.short.db"

    ctb.tiles(filename, ["x.18.169283"])
    # TODO: Make assertions about result
    # print("tiles:", tiles)
    # x = int(tiles[0]['xStart'])
    # fields = tiles[0]['fields']
    # TODO: Make assertions


def test_gene_annotations():
    runner = clt.CliRunner()
    input_file = op.join(testdir, "sample_data", "exon_unions_mm10.bed")
    f = tempfile.NamedTemporaryFile(delete=False)

    result = runner.invoke(
        cca.bedfile,
        [
            input_file,
            "--max-per-tile",
            "20",
            "--importance-column",
            "5",
            "--delimiter",
            "\t",
            "--assembly",
            "mm10",
            "--output-file",
            f.name,
        ],
    )

    # import traceback
    a, b, tb = result.exc_info
    """
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    """

    rows = ctb.tiles(f.name, ["x.0.0"])[0][1]
    assert len(rows) == 2

    rows = ctb.tiles(f.name, ["x.11.113"])[0][1]
    assert rows[0]["fields"][3] == "Lrp1b"

    rows = ctb.tiles(f.name, ["x.11.112"])[0][1]
    assert rows[0]["fields"][3] == "Lrp1b"


def test_random_importance():
    # check that when aggregating using random importance, all values that
    # are in a higher resolution tile are also in the lower resolution
    f = tempfile.NamedTemporaryFile(delete=False)

    runner = clt.CliRunner()
    input_file = op.join(testdir, "sample_data", "25435_PM15-000877_SM-7QK6O.seg")

    result = runner.invoke(
        cca.bedfile,
        [
            input_file,
            "--max-per-tile",
            "2",
            "--importance-column",
            "random",
            "--assembly",
            "b37",
            "--has-header",
            "--output-file",
            f.name,
        ],
    )

    # import traceback
    a, b, tb = result.exc_info
    """
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    """

    tsinfo = ctb.tileset_info(f.name)
    assert "version" in tsinfo
    assert len(tsinfo["header"]) > 0

    # check to make sure that tiles in the higher zoom levels
    # are all present in lower zoom levels
    found = {}

    for rect in ctb.tiles(f.name, ["x.5.15"])[0][1]:
        found[rect["xStart"]] = False

    for rect in ctb.tiles(f.name, ["x.6.30"])[0][1]:
        if rect["xStart"] in found:
            found[rect["xStart"]] = True

    for rect in ctb.tiles(f.name, ["x.6.31"])[0][1]:
        if rect["xStart"] in found:
            found[rect["xStart"]] = True

    for key, value in found.items():
        assert value

    pass


def test_no_chromosome_limit():
    f = tempfile.NamedTemporaryFile(delete=False)

    runner = clt.CliRunner()
    input_file = op.join(testdir, "sample_data", "geneAnnotationsExonsUnions.short.bed")

    result = runner.invoke(
        cca.bedfile,
        [
            input_file,
            "--max-per-tile",
            "60",
            "--importance-column",
            "5",
            "--assembly",
            "hg19",
            "--output-file",
            f.name,
        ],
    )

    # import traceback
    """
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    """
    a, b, tb = result.exc_info

    rows = ctb.tiles(f.name, ["x.0.0"])[0][1]
    foundOther = False

    for row in rows:
        if row["fields"][0] != "chr1":
            # print("row", row)
            assert row["xStart"] > 200000000
        if row["fields"][0] != "chr14":
            foundOther = True
        break
    # make sure there's chromosome other than 14 in the output
    assert foundOther

    os.remove(f.name)
    pass


def test_chromosome_limit():
    f = tempfile.NamedTemporaryFile(delete=False)

    runner = clt.CliRunner()
    input_file = op.join(testdir, "sample_data", "geneAnnotationsExonsUnions.short.bed")

    runner.invoke(
        cca.bedfile,
        [
            input_file,
            "--max-per-tile",
            "60",
            "--importance-column",
            "5",
            "--assembly",
            "hg19",
            "--chromosome",
            "chr14",
            "--output-file",
            f.name,
        ],
    )
    # TODO: Make assertions about result

    # print('output:', result.output, result)
    rows = ctb.tiles(f.name, ["x.0.0"])[0][1]

    for row in rows:
        assert row["fields"][0] == "chr14"

    os.remove(f.name)
    pass


def test_float_importance():
    f = tempfile.NamedTemporaryFile(delete=False)

    runner = clt.CliRunner()
    input_file = op.join(testdir, "sample_data", "test_float_importance.bed")

    runner.invoke(
        cca.bedfile,
        [
            input_file,
            "--max-per-tile",
            "2",
            "--importance-column",
            "4",
            "--assembly",
            "hg38",
            "--no-header",
            "--output-file",
            f.name,
        ],
    )
    # TODO: Make assertions about result



import math
import numpy as np
import clodius.tiles.format as ctf


def tiles_wrapper(array, tile_ids, not_nan_array=None):
    tile_values = []

    for tile_id in tile_ids:
        parts = tile_id.split(".")

        if len(parts) < 3:
            raise IndexError("Not enough tile info present")

        z = int(parts[1])
        x = int(parts[2])

        ret_array = tiles(array, z, x, not_nan_array).reshape((-1))

        tile_values += [(tile_id, ctf.format_dense_tile(ret_array))]

    return tile_values


def tileset_info(array, bounds=None, bins_per_dimension=1024):
    """
    Get the tileset info for the array
    """
    max_dim = max(array.shape)

    max_zoom = math.ceil(math.log(max_dim / bins_per_dimension) / math.log(2))
    max_zoom = 0 if max_zoom < 0 else max_zoom

    max_width = 2 ** max_zoom * bins_per_dimension
    # print('max_zoom:', max_zoom)

    scale_up = max_width / max_dim

    if bounds:
        min_pos = [bounds[0]]
        max_pos = [bounds[1]]

        # print('scale_up:', scale_up, max_pos[0] - min_pos[0])

        max_width = (max_pos[0] - min_pos[0]) * scale_up
    else:
        min_pos = [0]
        max_pos = [array.shape[0]]

    if len(array.shape) > 1:
        raise ValueError("The array shape is not a vector type", array.shape)
    return {
        "max_width": max_width,
        "min_pos": min_pos,
        "max_pos": max_pos,
        "max_zoom": max_zoom,
        "bins_per_dimension": bins_per_dimension,
        "tile_size": bins_per_dimension,
    }


def max_zoom_and_data_bounds(array, z, x, bin_size):
    """
    Return the maximum zoom level and data corresponding to the
    zoom level and position of the array.

    Parameters
    ----------
    array: np.array
        The array containing the raw numpy data
    z: int
        The zoom level
    x: int
        The x position
    """
    max_dim = array.shape[0]

    max_zoom = math.ceil(math.log(max_dim / bin_size) / math.log(2))
    max_zoom = 0 if max_zoom < 0 else max_zoom

    # max_width = 2 ** max_zoom * bin_size
    # print("max_width:", max_width, 'bin_size:', bin_size, 'max_zoom', max_zoom)

    tile_width = 2 ** (max_zoom - z) * bin_size

    x_start = x * tile_width
    x_end = min(array.shape[0], x_start + tile_width)

    # print("x_start:", x_start, x_end)

    return max_zoom, x_start, x_end


def tiles(array, z, x, not_nan_array=None, bin_size=1024):
    """
    Return tiles at the given positions.

    Parameters
    -----------
    array: np.array
        An nxn array containing values
    z: int
        The zoom level (0 corresponds to most zoomed out)
    x: int
        The x tile position
    not_nan_array: np.array
        An array storing the number of values which are not nan
        in the original array. Can be precalculated for speed.
    bin_size: int
        The number of values per bin
    """
    max_zoom, x_start, x_end = max_zoom_and_data_bounds(array, z, x, bin_size)
    data = array[x_start:x_end]

    num_to_sum = 2 ** (max_zoom - z)

    # add some data so that the data can be divided into squares
    divisible_x_width = num_to_sum * math.ceil(data.shape[0] / num_to_sum)
    divisible_x_pad = divisible_x_width - data.shape[0]

    a = np.pad(data, ((0, divisible_x_pad),), "constant", constant_values=(np.nan,))

    ret_array = np.nansum(a.reshape((-1, num_to_sum)), axis=1)

    if not_nan_array is None:
        not_nan_data = ~np.isnan(array[x_start:x_end])
    else:
        not_nan_data = not_nan_array[x_start:x_end]

    # we want to calculate the means of the data points
    na = np.pad(
        not_nan_data, ((0, divisible_x_pad)), "constant", constant_values=(np.nan,)
    )
    norm_array = np.nansum(na.reshape((-1, num_to_sum)), axis=1)
    ret_array = ret_array / (norm_array + 1)

    # determine how much to pad the array
    x_pad = bin_size - ret_array.shape[0]

    return np.pad(ret_array, ((0, x_pad)), "constant", constant_values=(np.nan,))



import numpy as np
import clodius.tiles.npvector as hgnv


def test_npvector():
    array = np.array(range(100))
    # print('ts:', hgnv.tileset_info(array))
    assert "max_width" in hgnv.tileset_info(array)

    hgnv.tiles(array, 0, 0)
    # TODO: Make assertions about tile returned.



import clodius.tiles.time_interval as hgti
import os.path as op


def test_tileset_info():
    filename = op.join("data", "sample_htime.json")

    hgti.tileset_info(filename)
    # TODO: Make assertions about info returned.



import math


def get_tile_box(zoom, x, y):
    """convert Google-style Mercator tile coordinate to
    (minlat, maxlat, minlng, maxlng) bounding box"""

    minlng, minlat = get_lng_lat_from_tile_pos(zoom, x, y)
    maxlng, maxlat = get_lng_lat_from_tile_pos(zoom, x + 1, y + 1)

    return (minlng, maxlng, minlat, maxlat)


def get_lng_lat_from_tile_pos(zoom, x, y):
    """convert Google-style Mercator tile coordinate to
    (lng, lat) of top-left corner of tile"""

    # "map-centric" latitude, in radians:
    lat_rad = math.pi - 2 * math.pi * y / (2 ** zoom)
    # true latitude:
    lat_rad = gudermannian(lat_rad)
    lat = lat_rad * 180.0 / math.pi

    # longitude maps linearly to map, so we simply scale:
    lng = -180.0 + 360.0 * x / (2 ** zoom)

    return (lng, lat)


def get_tile_pos_from_lng_lat(lng, lat, zoom):
    """convert lng/lat to Google-style Mercator tile coordinate (x, y)
    at the given zoom level"""

    lat_rad = lat * math.pi / 180.0
    # "map-centric" latitude, in radians:
    lat_rad = inv_gudermannian(lat_rad)

    x = 2 ** zoom * (lng + 180.0) / 360.0
    y = 2 ** zoom * (math.pi - lat_rad) / (2 * math.pi)

    return (x, y)


def gudermannian(x):
    return 2 * math.atan(math.exp(x)) - math.pi / 2


def inv_gudermannian(y):
    return math.log(math.tan((y + math.pi / 2) / 2))



import negspy.coordinates as nc


def load_chromsizes(chromsizes_filename, assembly=None):
    """
    Load a set of chromosomes from a file or using an assembly
    identifier. If using just an assembly identifier the chromsizes
    will be loaded from the negspy repository.

    Parameters:
    -----------
    chromsizes_filename: string
        The file containing the tab-delimited chromosome sizes
    assembly: string
        Assembly name (e.g. 'hg19'). Not necessary if a chromsizes_filename is passed in
    """
    if chromsizes_filename is not None:
        chrom_info = nc.get_chrominfo_from_file(chromsizes_filename)
        chrom_names = chrom_info.chrom_order
        chrom_sizes = [chrom_info.chrom_lengths[c] for c in chrom_info.chrom_order]
    else:
        if assembly is None:
            raise ValueError("No assembly or chromsizes specified")

        chrom_info = nc.get_chrominfo(assembly)
        chrom_names = nc.get_chromorder(assembly)
        chrom_sizes = nc.get_chromsizes(assembly)

    return (chrom_info, chrom_names, chrom_sizes)



import numpy as np


def tileset_info(f, bounds=None):
    if "min-pos" in f.attrs:
        min_pos = f.attrs["min-pos"]
    else:
        min_pos = [0, 0]

    if "max-pos" in f.attrs:
        max_pos = f.attrs["max-pos"]
    else:
        max_pos = f["resolutions"]["1"]["values"].shape

    return {
        "min_pos": min_pos,
        "max_pos": max_pos,
        "resolutions": [int(r) for r in f["resolutions"]],
        "mirror_tiles": "false",
        "bins_per_dimension": 256,
    }


def tiles(f, z, x, y):
    """
    Return tiles for the given region.

    Parameters:
    -----------
    f: h5py.File
        File pointer to the hdf5 file containing the matrices
    z: int
        The zoom level
    x: int
        The tile's x position
    y: int
        The tile's y position
    """
    resolutions = sorted(map(int, f["resolutions"].keys()))[::-1]
    tsinfo = tileset_info(f)
    n_bins = tsinfo["bins_per_dimension"]

    if z >= len(resolutions):
        raise ValueError("Zoom level out of bounds:", z, "resolutions:", resolutions)

    tile_width = tsinfo["bins_per_dimension"]

    # Where in the matrix the tile starts
    tile_x_start = x * tile_width
    tile_y_start = y * tile_width

    tile_x_end = tile_x_start + n_bins
    tile_y_end = tile_y_start + n_bins

    mat = f["resolutions"][str(resolutions[z])]["values"]
    data = mat[tile_y_start:tile_y_end, tile_x_start:tile_x_end]

    x_pad = n_bins - data.shape[0]
    y_pad = n_bins - data.shape[1]

    if x_pad > 0 or y_pad > 0:
        data = np.pad(
            data, ((0, x_pad), (0, y_pad)), "constant", constant_values=(np.nan, np.nan)
        )

    return data



import clodius.tiles.npmatrix as hgnp
import numpy as np


def test_numpy_matrix():
    grid = np.array(np.random.rand(100, 100))
    # print('grid:', grid)

    tile = hgnp.tiles(grid, 0, 0, 0)
    assert tile.shape == (256, 256)



import h5py
import clodius.tiles.hitile as hghi
import numpy as np
import os.path as op
import tempfile


def test_hitile():
    array_size = int(1e6)
    chunk_size = 2 ** 19

    data = np.random.random((array_size,))

    with tempfile.TemporaryDirectory() as td:
        output_file = op.join(td, "blah.hitile")
        hghi.array_to_hitile(data, output_file, zoom_step=6, chunks=(chunk_size,))

        with h5py.File(output_file, "r") as f:
            (means, mins, maxs) = hghi.get_data(f, 0, 0)
        # print("means, mins:", means[:10], mins[:10], maxs[:10])



from __future__ import print_function

import collections as col

try:
    import cStringIO as csio
except ImportError:
    import io as csio

import gzip
import itertools as it
import json
import numpy as np
import os
import os.path as op
import queue
import requests
import signal
import slugid
import sys
import time
import itertools
import traceback

from time import gmtime, strftime


def handle_exception(exc_type, exc_value, exc_traceback):
    print("".join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
    os._exit(1)


def tile_saver_worker(q, tile_saver, finished):
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    while not q.empty() or (not finished.value):
        # print "working...", q.qsize()
        try:
            (zoom_level, tile_pos, tile_bins) = q.get(timeout=1)
            # print("saving...", zoom_level, tile_pos, 'queue.qsize:', q.qsize(), q.empty(), "finished:", finished.value)
            tile_saver.save_binned_tile(zoom_level, tile_pos, tile_bins)
        except (KeyboardInterrupt, SystemExit):
            print("Exiting...")
            break
        except queue.Empty:
            tile_saver.flush()

    # print("finishing", q.qsize(), tile_saver)
    tile_saver.flush()


class TileSaver(object):
    def __init__(
        self,
        max_data_in_sparse,
        bins_per_dimension,
        num_dimensions,
        print_status=False,
        initial_value=0.0,
    ):
        self.max_data_in_sparse = max_data_in_sparse

        # self.max_data_in_sparse = 0
        self.bins_per_dimension = bins_per_dimension
        self.num_dimensions = num_dimensions
        self.print_status = print_status
        self.initial_value = initial_value

        pass

    def save_tile(self, tile):
        return

    def make_and_save_tile(self, zoom_level, tile_position, tile_data):
        # this implementation shouldn't do anything
        # derived classes should implement this functionality themselves

        tile_id = "{}.{}".format(zoom_level, ".".join(map(str, tile_position)))

        # print "saving:", tile_id
        tile = {"tile_id": tile_id, "tile_value": tile_data}

        self.save_tile(tile)

    def save_dense_tile(
        self, zoom_level, tile_position, tile_bins, min_value, max_value
    ):
        initial_values = [self.initial_value] * (
            self.bins_per_dimension ** self.num_dimensions
        )

        for (bin_pos, val) in tile_bins.items():
            index = int(
                sum([bp * self.bins_per_dimension ** i for i, bp in enumerate(bin_pos)])
            )
            initial_values[index] = val

        if len(self.initial_value) == 1:
            self.make_and_save_tile(
                zoom_level,
                tile_position,
                {
                    "dense": [round(v[0], 5) for v in initial_values],
                    "min_value": min_value,
                    "max_value": max_value,
                },
            )
        else:
            self.make_and_save_tile(
                zoom_level,
                tile_position,
                {
                    "dense": list(
                        it.chain.from_iterable(
                            [[round(y, 5) for y in v] for v in initial_values]
                        )
                    ),
                    "min_value": min_value,
                    "max_value": max_value,
                },
            )

    def save_sparse_tile(
        self, zoom_level, tile_position, tile_bins, min_value, max_value
    ):
        shown = []
        for (bin_pos, bin_val) in tile_bins.items():
            if len(bin_val) == 1:
                shown += [[list(map(float, bin_pos)), bin_val[0]]]
            else:
                shown += [[list(map(float, bin_pos)), list(bin_val)]]

        self.make_and_save_tile(
            zoom_level,
            tile_position,
            {"sparse": shown, "min_value": min_value, "max_value": max_value},
        )

    def save_tile_array(self, zoom_level, tile_position, tile_data):
        """
        Save a tile that has all of its data in one long array

        :param zoom_level: An integer zoom_level (0 for zoomed all the way out)
        :param tile_position: An n-dimensional array, where n is the number of dimensions
                              in the dataset.
        :param tile_data: The data in the tile.
        """
        min_value = min(tile_data)
        max_value = max(tile_data)

        self.make_and_save_tile(
            zoom_level,
            tile_position,
            {
                "dense": [round(v, 5) for v in tile_data],
                "min_value": min_value,
                "max_value": max_value,
            },
        )

    def save_binned_tile(self, zoom_level, tile_position, tile_bins):
        max_value = list(np.max(np.array(list(tile_bins.values())), axis=0))
        min_value = list(np.min(np.array(list(tile_bins.values())), axis=0))

        if len(tile_bins) < self.max_data_in_sparse:
            self.save_sparse_tile(
                zoom_level,
                tile_position,
                tile_bins,
                min_value=min_value,
                max_value=max_value,
            )
        else:
            self.save_dense_tile(
                zoom_level,
                tile_position,
                tile_bins,
                min_value=min_value,
                max_value=max_value,
            )

    def flush():
        return


class EmptyTileSaver(TileSaver):
    def __init__(self, max_data_in_sparse, bins_per_dimension, num_dimensions):
        super(EmptyTileSaver, self).__init__(
            max_data_in_sparse, bins_per_dimension, num_dimensions
        )


class ColumnFileTileSaver(TileSaver):
    def __init__(
        self,
        max_data_in_sparse,
        bins_per_dimension,
        num_dimensions,
        file_path,
        log_file,
        print_status,
        initial_value,
    ):
        super(ColumnFileTileSaver, self).__init__(
            max_data_in_sparse,
            bins_per_dimension,
            num_dimensions,
            print_status,
            initial_value,
        )
        self.file_path = file_path
        self.bulk_txt = csio.StringIO()
        self.bulk_txt_len = 0
        self.log_file = log_file

    def save_tile(self, val):
        """
        if ('dense' in val['tile_value']):
            value_pos = col.defaultdict(list)
            dense_values = val['tile_value']['dense']
            dense_values = [(x,len(list(y))) for (x,y) in it.groupby(dense_values)]
            dense_values = [item for sublist in dense_values for item in sublist]
            val['tile_value']['dense'] = dense_values
            for i,value in enumerate(dense_values):
                value_pos[value] += [i]
            for key in value_pos:
                sorted_value_pos = sorted(value_pos[key])
                diffs = []
                diffs += [sorted_value_pos[0]]
                for i in range(len(sorted_value_pos)-1):
                    diffs += [sorted_value_pos[i+1] - sorted_value_pos[i]]

                value_pos[key] = diffs
            val['tile_value']['dense'] = value_pos.items()
        """

        """
        if ('sparse' in val['tile_value']):
            sparse_values = val['tile_value']['sparse']
            value_pos = col.defaultdict(list)
            for sparse_value in sparse_values:
                value_pos[sparse_value[1]] += [sparse_value[0]]
            val['tile_value']['sparse'] = value_pos.items()

            value_xs_ys = []
            for value, poss in value_pos.items():
                poss = sorted(poss)
                xs = [p[0] for p in poss]
                ys = [p[1] for p in poss]
                value_xs_ys += [value, xs, ys]
            val['tile_value']['sparse'] = value_xs_ys
        """

        # [[1.0, [[78.0, 123.0], [64.0, 153.0]]]]

        if val["tile_id"] == "tileset_info":
            self.bulk_txt.write(val["tile_id"] + "\t" + "1" + "\t" + "1" + "\t")
        else:
            ti = val["tile_id"].split(".")
            self.bulk_txt.write(
                str(int(ti[0]) + 1)
                + "\t"
                + str(int(ti[1]) + 1)
                + "\t"
                + str(int(ti[1]) + 1)
                + "\t"
            )

        self.bulk_txt.write(json.dumps(val) + "\n")
        curr_pos = self.bulk_txt.tell()
        # print "curr_pos:", curr_pos,self.bulk_txt.getvalue()
        # self.bulk_txt.write(new_string)
        if curr_pos > 2000000:
            self.flush()

    def flush(self):
        if self.bulk_txt.tell() > 0:
            try:
                with open(self.file_path, "a") as column_file:
                    column_file.write(self.bulk_txt.getvalue())
            except Exception as ex:
                if self.log_file is not None:
                    with open(self.log_file, "a") as f:
                        f.write(strftime("%Y-%m-%d %H:%M:%S", gmtime()))
                        f.write(ex)

        self.bulk_txt_len = 0
        self.bulk_txt.close()
        self.bulk_txt = csio.StringIO()


class ElasticSearchTileSaver(TileSaver):
    def __init__(
        self,
        max_data_in_sparse=None,
        bins_per_dimension=None,
        num_dimensions=None,
        es_path=None,
        log_file=None,
        print_status=False,
        initial_value=None,
    ):
        super(ElasticSearchTileSaver, self).__init__(
            max_data_in_sparse,
            bins_per_dimension,
            num_dimensions,
            print_status,
            initial_value,
        )
        self.es_path = es_path
        self.bulk_txt = csio.StringIO()
        self.bulk_txt_len = 0
        self.log_file = log_file

    def save_tile(self, val):
        # this implementation shouldn't do anything
        # derived classes should implement this functionality themselves

        # self.bulk_txt.write(json.dumps({"index": {"_id": val['tile_id']}}) + "\n")
        if "sparse" in val["tile_value"]:
            sparse_values = val["tile_value"]["sparse"]

            value_pos = col.defaultdict(list)
            for sparse_value in sparse_values:
                if len(self.initial_value) != 1:
                    value_pos[tuple(sparse_value[1])] += [sparse_value[0]]
                else:
                    value_pos[sparse_value[1]] += [sparse_value[0]]
            # val['tile_value']['sparse'] = value_pos.items()

            value_xs_ys = []
            for value, poss in value_pos.items():
                # sparse values are stored as the following:
                # value, # of positions it's found in, list of the positions
                poss = sorted(poss)
                if len(self.initial_value) == 1:
                    value_xs_ys += [float(value)]
                else:
                    value_xs_ys += list(value)
                value_xs_ys += [float(len(poss))]

                for i in range(len(poss[0])):
                    value_xs_ys += [p[i] for p in poss]

            val["tile_value"]["sparse"] = value_xs_ys

        self.save_value(val["tile_id"], val)

    def save_value(self, doc_id, doc):
        """
        if ('dense' in val['tile_value']):
            print val['tile_id'], len([x for x in val['tile_value']['dense'] if x > 0])
        """

        # val['tile_value']['dense'] = []

        self.bulk_txt.write('{{"index": {{"_id": "{}"}}}}\n'.format(doc_id))
        self.bulk_txt.write(json.dumps(doc) + "\n")

        """
        self.bulk_txt.write('{{"tile_id": {}, "tile_value": '.format(val['tile_id']))

        if 'sparse' in val['tile_value']:
            self.bulk_txt.write(' {{ "sparse": [ ')
            self.bulk_txt.write(','.join(
                ['{{ "pos": [{}], "value": {} }}'.format(
                    ",".join([str(y) for y in x['pos']]), x['value'])
                    for x in val['tile_value']['sparse']]))
            self.bulk_txt.write('] }}')


        # sys.exit(1)
        # new_string += str(val) + "\n"

        # self.bulk_txt_len += len(new_string)
        """

        curr_pos = self.bulk_txt.tell()
        # self.bulk_txt.write(new_string)
        if curr_pos > 5000000:
            self.flush()

    def flush(self):
        if self.bulk_txt.tell() > 0:
            # only save the tile if it had enough data
            try:
                save_to_elasticsearch(
                    "http://" + self.es_path + "/_bulk",
                    self.bulk_txt.getvalue(),
                    self.print_status,
                )
            except Exception as ex:
                if self.log_file is not None:
                    with open(self.log_file, "a") as f:
                        f.write(strftime("%Y-%m-%d %H:%M:%S", gmtime()))
                        f.write(ex)

            self.bulk_txt_len = 0
            self.bulk_txt.close()
            self.bulk_txt = csio.StringIO()


def save_tile_to_elasticsearch(
    partition, elasticsearch_nodes, elasticsearch_path, print_status=False
):
    bulk_txt = ""
    es_url = op.join(elasticsearch_nodes, elasticsearch_path)
    put_url = op.join(es_url, "_bulk")

    for val in partition:
        bulk_txt += json.dumps({"index": {"_id": val["tile_id"]}}) + "\n"
        bulk_txt += json.dumps(val) + "\n"

        if len(bulk_txt) > 5000000:
            save_to_elasticsearch("http://" + put_url, bulk_txt, print_status)
            bulk_txt = ""

    print("bulk_txt:", bulk_txt)
    if len(bulk_txt) > 0:
        save_to_elasticsearch("http://" + put_url, bulk_txt, print_status)


def save_to_elasticsearch(url, data, print_status=False):
    """
    Save some data to elastic search.

    The data should be a string suitable for bulk import by
    elasticsearch. The url should be the location of the index, document
    type, along with the _bulk destination.

    :param url: The elasticsearch url that will ingest the data
                e.g. localhost:9200/hg19/tiles/_bulk
    :param data: The data to import.
                e.g. {"index": {"_id": "blah", "my_json": {"x": 2}}}
    """
    saved = False
    to_sleep = 1

    uid = slugid.nice()
    # print("print_status:", print_status)
    while not saved:
        try:
            r = requests.post(url, data=data, timeout=8)
            if print_status:
                print("\nSaved", uid, r, "len(data):", len(data), url)
                # print("data:", data)
            saved = True
            # print "data:", data
        except Exception as ex:

            to_sleep *= 2
            print(
                "\nError saving to elastic search (",
                uid,
                "), sleeping:",
                to_sleep,
                ex,
                file=sys.stderr,
            )
            time.sleep(to_sleep)

            if to_sleep > 600:
                with open("unsaved.err", "a") as f:
                    f.write("UNSAVED url:", url, "\n")
                    f.write(data)
                    f.flush()
                print("Slept too long, returning", file=sys.stderr)
                raise


def save_tile(tile, output_dir, gzip_output):
    """
    Save a tile to a particular base directory.

    This function create the appropriate sub-directory based on the
    key.

    They key should be in the format (zoom_level, pos1, pos2...)
    e.g. (5,4,5)
    """
    key = tile[0]
    tile_value = tile[1]

    outpath = op.join(output_dir, ".".join(map(str, key)))
    outdir = op.dirname(outpath)

    if not op.exists(outdir):
        try:
            os.makedirs(outdir)
        except OSError as oe:
            # somebody probably made the directory in between when we
            # checked if it exists and when we're making it
            print("Error making directories:", oe, file=sys.stderr)

    output_json = {
        "_source": {"tile_id": ".".join(map(str, key)), "tile_value": tile_value}
    }
    if gzip_output:
        with gzip.open(outpath + ".gz", "w") as f:
            f.write(json.dumps(output_json, indent=2))
    else:
        with open(outpath, "w") as f:
            f.write(json.dumps(output_json, indent=2))



from __future__ import print_function

import clodius.tiles.bam as ctb
import os.path as op
import unittest


class MyTestCase(unittest.TestCase):
    def test_tileset_info(self):
        filename_matched = op.join("data", "SRR1770413.sorted.short.bam")

        filename_mismatched = op.join("data", "SRR1770413.mismatched_bai.bam")

        tsinfo = ctb.tileset_info(filename_matched)
        assert "max_zoom" in tsinfo

        tsinfo = ctb.tileset_info(filename_mismatched)
        assert "max_zoom" in tsinfo

    def test_tiles(self):
        filename_matched = op.join("data", "SRR1770413.sorted.short.bam")

        filename_mismatched = op.join("data", "SRR1770413.mismatched_bai.bam")

        index_filename = op.join("data", "SRR1770413.different_index_filename.bai")

        tile = ctb.tiles(filename_matched, ["x.9.0"])

        assert len(tile) > 0

        # missing index
        self.assertRaises(ValueError, ctb.tiles, filename_mismatched, ["x.9.0"])

        tile = ctb.tiles(filename_mismatched, ["x.9.0"], index_filename=index_filename)

        assert len(tile) > 0
        # example tile: [('x.9.0', { id: [....]})]
        assert len(tile[0][1]["id"]) > 10

        tile = ctb.tiles(
            filename_mismatched,
            ["x.9.0"],
            index_filename=index_filename,
            max_tile_width=10,
        )

        assert "error" in tile[0][1]



import json
import math
import os
import sqlite3
import collections as col


def get_tile_box(zoom, x, y):
    """convert Google-style Mercator tile coordinate to
    (minlat, maxlat, minlng, maxlng) bounding box"""

    minlng, minlat = get_lng_lat_from_tile_pos(zoom, x, y)
    maxlng, maxlat = get_lng_lat_from_tile_pos(zoom, x + 1, y + 1)

    return (minlng, maxlng, minlat, maxlat)


def get_lng_lat_from_tile_pos(zoom, x, y):
    """convert Google-style Mercator tile coordinate to
    (lng, lat) of top-left corner of tile"""

    # "map-centric" latitude, in radians:
    lat_rad = math.pi - 2 * math.pi * y / (2 ** zoom)
    # true latitude:
    lat_rad = gudermannian(lat_rad)
    lat = lat_rad * 180.0 / math.pi

    # longitude maps linearly to map, so we simply scale:
    lng = -180.0 + 360.0 * x / (2 ** zoom)

    return (lng, lat)


def get_tile_pos_from_lng_lat(lng, lat, zoom):
    """convert lng/lat to Google-style Mercator tile coordinate (x, y)
    at the given zoom level"""

    lat_rad = lat * math.pi / 180.0
    # "map-centric" latitude, in radians:
    lat_rad = inv_gudermannian(lat_rad)

    x = 2 ** zoom * (lng + 180.0) / 360.0
    y = 2 ** zoom * (math.pi - lat_rad) / (2 * math.pi)

    return (x, y)


def gudermannian(x):
    return 2 * math.atan(math.exp(x)) - math.pi / 2


def inv_gudermannian(y):
    return math.log(math.tan((y + math.pi / 2) / 2))


def tileset_info(filepath):
    if not os.path.isfile(filepath):
        return {"error": "Tileset info is not available!"}

    db = sqlite3.connect(filepath)

    res = db.execute("SELECT * FROM tileset_info").fetchone()

    o = {
        "zoom_step": res[0],
        "tile_size": res[1],
        "max_zoom": res[2],
        "min_pos": [res[3], res[5]],
        "max_pos": [res[4], res[6]],
        "max_data_length": res[1] * 2 ** res[2],
    }

    return o


def get_tiles(db_file, zoom, x, y, width=1, height=1):
    """
    Retrieve a contiguous set of tiles from a 2D db tile file.

    Parameters
    ----------
    db_file: str
        The filename of the sqlite db file
    zoom: int
        The zoom level
    x: int
        The x position of the first tile
    y: int
        The y position of the first tile
    width: int
        The width of the block of tiles to retrieve
    height: int
        The height of the block of tiles to retrieve

    Returns
    -------
    tiles: {pos: tile_value}
        A set of tiles, indexed by position
    """
    conn = sqlite3.connect(db_file)

    c = conn.cursor()

    lng_from, _, lat_from, _ = get_tile_box(zoom, x, y)
    _, lng_to, _, lat_to = get_tile_box(zoom, x + width - 1, y + height - 1)

    # Note the range query is being done in lng-lat. Since the coords go from:
    # - Longitude: -180 to 180
    # - Latidue: 90 to -90 (NOTE THE SWAP HERE!!!)
    # we are indexing min and max longitude and latitude but for querying we
    # are using from and to longitude and latitude. Hence, the comparator and
    # min/max for latitude are flipped (max == from lat; min == to lat)
    query = """
    SELECT
        minLng, maxLng, maxLat, minLat, uid, importance, geometry, properties,
        intervals.id
    FROM
        intervals, position_index
    WHERE
        intervals.id=position_index.id AND
        zoomLevel <= ? AND
        rMaxLng >= ? AND
        rMinLng <= ? AND
        rMinLat <= ? AND
        rMaxLat >= ?
    """

    rows = c.execute(query, (zoom, lng_from, lng_to, lat_from, lat_to)).fetchall()

    new_rows = col.defaultdict(list)

    for i in range(x, x + width):
        for j in range(y, y + height):
            new_rows[(i, j)] = []

    for r in rows:
        try:
            uid = r[4].decode("utf-8")
        except AttributeError:
            uid = r[4]
        try:
            id = r[8].decode("utf-8")
        except AttributeError:
            id = r[8]

        x_start, y_start = get_tile_pos_from_lng_lat(r[0], r[2], zoom)
        x_end, y_end = get_tile_pos_from_lng_lat(r[1], r[3], zoom)

        try:
            geometry = json.loads(r[6])
        except Exception:
            geometry = None
            pass

        try:
            properties = json.loads(r[7])
        except Exception:
            properties = None
            pass

        for i in range(x, x + width):
            for j in range(y, y + height):
                # Add annotations to each tile in which they are visible
                if x_start < i + 1 and x_end >= i and y_start < j + 1 and y_end >= j:
                    # add the position offset to the returned values
                    new_rows[(i, j)] += [
                        {
                            "xStart": r[0],
                            "xEnd": r[1],
                            "yStart": r[2],
                            "yEnd": r[3],
                            "importance": r[5],
                            "uid": uid,
                            "geometry": geometry,
                            "properties": properties,
                            "id": id,
                        }
                    ]
    conn.close()

    return new_rows



import collections as col
import sqlite3


def get_2d_tileset_info(db_file):
    conn = sqlite3.connect(db_file)
    c = conn.cursor()

    row = c.execute("SELECT * from tileset_info").fetchone()
    tileset_info = {
        "zoom_step": row[0],
        "max_length": row[1],
        "assembly": row[2],
        "chrom_names": row[3],
        "chrom_sizes": row[4],
        "tile_size": row[5],
        "max_zoom": row[6],
        "max_width": row[7],
        "min_pos": [1, 1],
        "max_pos": [row[1], row[1]],
    }
    conn.close()

    return tileset_info


def get_1D_tiles(db_file, zoom, tile_x_pos, numx):
    """
    Retrieve a contiguous set of tiles from a 2D db tile file.

    Parameters
    ----------
    db_file: str
        The filename of the sqlite db file
    zoom: int
        The zoom level
    tile_x_pos: int
        The x position of the first tile
    numx: int
        The width of the block of tiles to retrieve

    Returns
    -------
    tiles: {pos: tile_value}
        A set of tiles, indexed by position
    """
    tileset_info = get_2d_tileset_info(db_file)

    conn = sqlite3.connect(db_file)

    c = conn.cursor()
    tile_width = tileset_info["max_width"] / 2 ** zoom

    tile_x_start_pos = tile_width * tile_x_pos
    tile_x_end_pos = tile_x_start_pos + (numx * tile_width)

    # print('tile_x_start:', tile_x_start_pos, tile_x_end_pos)

    query = """
    SELECT fromX, toX, fromY, toY, chrOffset, importance, fields, uid
    FROM intervals,position_index
    WHERE
        intervals.id=position_index.id AND
        zoomLevel <= {} AND
        rToX >= {} AND
        rFromX <= {}
    """.format(
        zoom, tile_x_start_pos, tile_x_end_pos
    )

    rows = c.execute(query).fetchall()

    new_rows = col.defaultdict(list)
    # print("len(rows)", len(rows))

    for r in rows:
        try:
            uid = r[7].decode("utf-8")
        except AttributeError:
            uid = r[7]

        x_start = r[0]
        x_end = r[1]
        y_start = r[2]
        y_end = r[3]

        for i in range(tile_x_pos, tile_x_pos + numx):
            tile_x_start = i * tile_width
            tile_x_end = (i + 1) * tile_width

            if x_start < tile_x_end and x_end >= tile_x_start:
                # add the position offset to the returned values
                new_rows[i] += [
                    {
                        "xStart": x_start,
                        "xEnd": x_end,
                        "yStart": y_start,
                        "yEnd": y_end,
                        "chrOffset": r[4],
                        "importance": r[5],
                        "uid": uid,
                        "fields": r[6].split("\t"),
                    }
                ]
    conn.close()

    return new_rows


def get_2D_tiles(db_file, zoom, tile_x_pos, tile_y_pos, numx=1, numy=1):
    """
    Retrieve a contiguous set of tiles from a 2D db tile file.

    Parameters
    ----------
    db_file: str
        The filename of the sqlite db file
    zoom: int
        The zoom level
    tile_x_pos: int
        The x position of the first tile
    tile_y_pos: int
        The y position of the first tile
    numx: int
        The width of the block of tiles to retrieve
    numy: int
        The height of the block of tiles to retrieve

    Returns
    -------
    tiles: {pos: tile_value}
        A set of tiles, indexed by position
    """
    tileset_info = get_2d_tileset_info(db_file)

    conn = sqlite3.connect(db_file)

    c = conn.cursor()
    tile_width = tileset_info["max_width"] / 2 ** zoom

    tile_x_start_pos = tile_width * tile_x_pos
    tile_x_end_pos = tile_x_start_pos + (numx * tile_width)

    tile_y_start_pos = tile_width * tile_y_pos
    tile_y_end_pos = tile_y_start_pos + (numy * tile_width)

    query = """
    SELECT fromX, toX, fromY, toY, chrOffset, importance, fields, uid
    FROM intervals,position_index
    WHERE
        intervals.id=position_index.id AND
        zoomLevel <= {} AND
        rToX >= {} AND
        rFromX <= {} AND
        rToY >= {} AND
        rFromY <= {}
    """.format(
        zoom, tile_x_start_pos, tile_x_end_pos, tile_y_start_pos, tile_y_end_pos
    )

    rows = c.execute(query).fetchall()

    new_rows = col.defaultdict(list)

    for r in rows:
        try:
            uid = r[7].decode("utf-8")
        except AttributeError:
            uid = r[7]

        x_start = r[0]
        x_end = r[1]
        y_start = r[2]
        y_end = r[3]

        for i in range(tile_x_pos, tile_x_pos + numx):
            for j in range(tile_y_pos, tile_y_pos + numy):
                tile_x_start = i * tile_width
                tile_x_end = (i + 1) * tile_width

                tile_y_start = j * tile_width
                tile_y_end = (j + 1) * tile_width

                if (
                    x_start < tile_x_end
                    and x_end >= tile_x_start
                    and y_start < tile_y_end
                    and y_end >= tile_y_start
                ):
                    # add the position offset to the returned values
                    new_rows[(i, j)] += [
                        {
                            "xStart": r[0],
                            "xEnd": r[1],
                            "yStart": r[2],
                            "yEnd": r[3],
                            "chrOffset": r[4],
                            "importance": r[5],
                            "uid": uid,
                            "fields": r[6].split("\t"),
                        }
                    ]
    conn.close()

    return new_rows



import numpy as np
import clodius.tiles.npvector as ctn


def tiles(labels, z, x, importances=None, tile_size=16):
    """
    Return tiles from the array. If importances are provided,
    then they will be used to prioritize entries. Otherwise we'll
    assign random importances to the entire array.
    """
    max_zoom, x_start, x_end = ctn.max_zoom_and_data_bounds(labels, z, x, tile_size)

    tile_labels = np.asarray(labels[x_start:x_end])
    if tile_labels.dtype.kind == "S":
        tile_labels = tile_labels.astype("U")

    if importances is not None and len(importances):
        tile_importances = np.asarray(importances[x_start:x_end])
        indices = np.argsort(tile_importances)[::-1][:tile_size]
    else:
        tile_importances = np.zeros(x_end - x_start)
        indices = np.linspace(x_start, x_end - 1, tile_size, dtype=int) - x_start

    return [
        {"x": x, "label": label, "importance": importance}
        for x, label, importance in zip(
            (x_start + indices).tolist(),
            tile_labels[indices].tolist(),
            tile_importances[indices].tolist(),
        )
    ]


def tiles_wrapper(array, tile_ids, importances):
    tile_values = []

    for tile_id in tile_ids:
        parts = tile_id.split(".")

        if len(parts) < 3:
            raise IndexError("Not enough tile info present")

        z = int(parts[1])
        x = int(parts[2])

        ret_array = tiles(array, z, x, importances)

        tile_values += [(tile_id, ret_array)]

    return tile_values



from __future__ import print_function

import gzip
import h5py
import math
import numpy as np
import logging
import os
import os.path as op
import sys

logger = logging.getLogger(__name__)


def bedfile_to_multivec(
    input_filenames,
    f_out,
    bedline_to_chrom_start_end_vector,
    base_resolution,
    has_header,
    chunk_size,
    num_rows,
    row_infos=None,
):
    """
    Convert an epilogos bedfile to multivec format.
    """

    files = []
    for input_filename in input_filenames:
        if op.splitext(input_filename)[1] == ".gz":
            files += [gzip.open(input_filename, "rt")]
        else:
            files += [open(input_filename, "r")]

    FILL_VALUE = np.nan

    # batch regions because h5py is really bad at writing
    batch_length = chunk_size
    batch = []

    # the current index in the dataset
    curr_index = 0
    # the start of the batch in the dataset
    batch_start_index = 0

    if has_header:
        files[0].readline()

    prev_chrom = None

    print("base_resolution:", base_resolution)
    for _, lines in enumerate(zip(*files)):

        # Identifies bedfile headers and ignore them
        if lines[0].startswith("browser") or lines[0].startswith("track"):
            continue

        chrom, start, end, vector = bedline_to_chrom_start_end_vector(lines, row_infos)
        # if vector[0] > 0 or vector[1] > 0:
        if len(vector) < len(lines) * num_rows:
            logger.warn("Lines contain fewer columns than expected: %s", lines)
            vector += [np.nan] * (len(lines) * num_rows - len(vector))

        if start % base_resolution != 0:
            logger.error(
                "The start coordinate is not a multiple of the resolution in line: %s",
                lines,
            )
            sys.exit(1)

        if prev_chrom is not None and chrom != prev_chrom:
            # we've reached a new chromosome so we'll dump all
            # the previous values
            # print("len(batch:", len(batch),
            #       "batch_start_index", batch_start_index)
            f_out[prev_chrom][
                batch_start_index : batch_start_index + len(batch)
            ] = np.array(batch)

            # we're starting a new chromosome so we start from the beginning
            curr_index = 0
            batch_start_index = 0
            batch = []

        prev_chrom = chrom

        # print('parts', parts)
        # print('chrom:', chrom, start)

        data_start_index = start // base_resolution

        # if the bedfile skips over a region, we have to add it as empty values
        # to preserve our batch writing
        while curr_index < data_start_index:
            batch += [[FILL_VALUE] * len(vector)]
            curr_index += 1

        """
        if curr_index != data_start_index:
            print("curr_index:", curr_index, data_start_index)
            print("line:", line)
        """

        assert curr_index == data_start_index
        # print('vector', vector)

        # When the binsize is not equal to the base_resolution
        # "break down" the binsize into bins of the rbase_esolution size
        # and add the values to each bin.
        # If there is a remainder, add an additional bin
        data_end_index = math.ceil(end / base_resolution)

        while curr_index < data_end_index:
            batch += [vector]
            curr_index += 1

        # fill in empty

        if len(batch) >= batch_length:
            # dump batch
            try:
                f_out[chrom][
                    batch_start_index : batch_start_index + len(batch)
                ] = np.array(batch)
            except TypeError as ex:
                print("Error:", ex, file=sys.stderr)
                print("Probably need to set the --num-rows parameter", file=sys.stderr)
                return

            batch = []
            batch_start_index = curr_index
            print("dumping batch:", chrom, batch_start_index)

    f_out[chrom][batch_start_index : batch_start_index + len(batch)] = np.array(batch)


def create_multivec_multires(
    array_data,
    chromsizes,
    agg,
    starting_resolution=1,
    tile_size=1024,
    output_file="/tmp/my_file.multires",
    row_infos=None,
):
    """
    Create a multires file containing the array data
    aggregated at multiple resolutions.

    Parameters
    ----------
    array_data: {'chrom_key': np.array, }
        The array data to aggregate organized by chromosome
    chromsizes: [('chrom_key', size),...]
    agg: lambda
        The function that will aggregate the data. Should
        take an array as input and create another array of
        roughly half the length
    starting_resolution: int (default 1)
        The starting resolution of the input data
    tile_size: int
        The tile size that we want higlass to use. This should
        depend on the size of the data but when in doubt, just use
        256.
    """
    filename = output_file

    # this is just so we can run this code
    # multiple times without h5py complaining
    if op.exists(filename):
        os.remove(filename)

    # this will be the file that contains our multires data
    f = h5py.File(filename, "w")

    # store some metadata
    f.create_group("info")
    f["info"].attrs["tile-size"] = tile_size

    f.create_group("resolutions")
    f.create_group("chroms")

    # start with a resolution of 1 element per pixel
    curr_resolution = starting_resolution

    # this will be our sample highest-resolution array
    # and it will be stored under the resolutions['1']
    # dataset
    f["resolutions"].create_group(str(curr_resolution))

    chroms, lengths = zip(*chromsizes)
    chrom_array = np.array(chroms, dtype="S")

    # row_infos = None
    if "row_infos" in array_data.attrs:
        row_infos = array_data.attrs["row_infos"]

    # add the chromosome information
    if row_infos is not None:
        f["resolutions"][str(curr_resolution)].attrs.create("row_infos", row_infos)

    f["resolutions"][str(curr_resolution)].create_group("chroms")
    f["resolutions"][str(curr_resolution)].create_group("values")
    f["resolutions"][str(curr_resolution)]["chroms"].create_dataset(
        "name",
        shape=(len(chroms),),
        dtype=chrom_array.dtype,
        data=chrom_array,
        compression="gzip",
    )
    f["resolutions"][str(curr_resolution)]["chroms"].create_dataset(
        "length", shape=(len(chroms),), data=lengths, compression="gzip"
    )

    f["chroms"].create_dataset(
        "name",
        shape=(len(chroms),),
        dtype=chrom_array.dtype,
        data=chrom_array,
        compression="gzip",
    )
    f["chroms"].create_dataset(
        "length", shape=(len(chroms),), data=lengths, compression="gzip"
    )

    # add the data
    for chrom, length in zip(chroms, lengths):
        if chrom not in array_data:
            print("Missing chrom {} in input file".format(chrom), file=sys.stderr)
            continue

        # print("creating new dataset")
        f["resolutions"][str(curr_resolution)]["values"].create_dataset(
            str(chrom), array_data[chrom].shape, compression="gzip"
        )
        standard_chunk_size = 1e5
        start = 0
        chrom_data = f["resolutions"][str(curr_resolution)]["values"][chrom]

        chunk_size = int(min(standard_chunk_size, len(chrom_data)))
        # print("array_data.shape", array_data[chrom].shape)

        while start < len(chrom_data):
            # see above section
            chrom_data[start : start + chunk_size] = array_data[chrom][
                start : start + chunk_size
            ]
            start += int(min(standard_chunk_size, len(array_data[chrom]) - start))

    # the maximum zoom level corresponds to the number of aggregations
    # that need to be performed so that the entire extent of
    # the dataset fits into one tile
    total_length = sum(lengths)
    # print("total_length:", total_length, "tile_size:", tile_size, "starting_resolution:", starting_resolution)
    max_zoom = math.ceil(
        math.log(total_length / (tile_size * starting_resolution)) / math.log(2)
    )

    # we're going to go through and create the data for the different
    # zoom levels by summing adjacent data points
    prev_resolution = curr_resolution

    for i in range(max_zoom):
        # each subsequent zoom level will have half as much data
        # as the previous
        curr_resolution = prev_resolution * 2
        f["resolutions"].create_group(str(curr_resolution))

        # add information about each of the rows
        if row_infos is not None:
            f["resolutions"][str(curr_resolution)].attrs.create("row_infos", row_infos)

        f["resolutions"][str(curr_resolution)].create_group("chroms")
        f["resolutions"][str(curr_resolution)].create_group("values")
        f["resolutions"][str(curr_resolution)]["chroms"].create_dataset(
            "name",
            shape=(len(chroms),),
            dtype=chrom_array.dtype,
            data=chrom_array,
            compression="gzip",
        )
        f["resolutions"][str(curr_resolution)]["chroms"].create_dataset(
            "length", shape=(len(chroms),), data=lengths, compression="gzip"
        )

        for chrom, length in zip(chroms, lengths):
            if chrom not in f["resolutions"][str(prev_resolution)]["values"]:
                continue

            # next_level_length = math.ceil(
            #     len(f['resolutions'][str(prev_resolution)]['values'][chrom]) / 2)

            start = 0

            chrom_data = f["resolutions"][str(prev_resolution)]["values"][chrom]

            standard_chunk_size = 1e5
            chunk_size = int(min(standard_chunk_size, len(chrom_data)))

            new_shape = list(chrom_data.shape)
            new_shape[0] = math.ceil(new_shape[0] / 2)
            new_shape = tuple(new_shape)

            f["resolutions"][str(curr_resolution)]["values"].create_dataset(
                chrom, new_shape, compression="gzip"
            )

            while start < len(chrom_data):
                old_data = f["resolutions"][str(prev_resolution)]["values"][chrom][
                    start : start + chunk_size
                ]
                # print("prev_resolution:", prev_resolution)
                # print("old_data.shape", old_data.shape)

                # this is a sort of roundabout way of calculating the
                # shape of the aggregated array, but all its doing is
                # just halving the first dimension of the previous shape
                # without taking into account the other dimensions
                # print("11 old_data.shape", old_data.shape)
                if len(old_data) % 2 != 0:
                    # we need our array to have an even number of elements
                    # so we just add the last element again
                    old_data = np.concatenate((old_data, [old_data[-1]]))
                    chunk_size += 1
                # print("22 old_data.shape", old_data.shape)

                # print('old_data:', old_data)
                # print("shape:", old_data.shape)
                # actually sum the adjacent elements
                # print("old_data.shape", old_data.shape)
                new_data = agg(old_data)

                """
                print("zoom_level:", max_zoom - 1 - i,
                      "resolution:", curr_resolution,
                      "new_data length", len(new_data))
                """
                f["resolutions"][str(curr_resolution)]["values"][chrom][
                    int(start / 2) : int(start / 2 + chunk_size / 2)
                ] = new_data
                start += int(min(standard_chunk_size, len(chrom_data) - start))

        prev_resolution = curr_resolution
    return f



import unittest

import numpy as np
from numpy.testing import assert_array_equal

from clodius.tiles.mrmatrix import tileset_info, tiles


class MockHdf5(dict):
    # By wrapping a dict in our own class, we can add arbitrary attributes.
    pass


class TilesetInfoTest(unittest.TestCase):
    def setUp(self):
        tileset_stub = {"resolutions": {"1": {"values": np.array([[1, 2], [3, 4]])}}}

        self.tileset = MockHdf5(tileset_stub)
        self.tileset.attrs = {}

        self.tileset_min = MockHdf5(tileset_stub)
        self.tileset_min.attrs = {"min-pos": (1, 1)}

        self.tileset_max = MockHdf5(tileset_stub)
        self.tileset_max.attrs = {"max-pos": (9, 9)}

        self.info = {
            "bins_per_dimension": 256,
            "max_pos": (2, 2),
            "min_pos": [0, 0],
            "mirror_tiles": "false",
            "resolutions": [1],
        }

    def test_default(self):
        self.assertEqual(tileset_info(self.tileset), self.info)
        self.assertEqual(tileset_info(self.tileset, bounds="???"), self.info)

    def test_with_min(self):
        self.info["min_pos"] = (1, 1)
        self.assertEqual(tileset_info(self.tileset_min), self.info)
        self.assertEqual(tileset_info(self.tileset_min, bounds="???"), self.info)

    def test_with_max(self):
        self.info["max_pos"] = (9, 9)
        self.assertEqual(tileset_info(self.tileset_max), self.info)
        self.assertEqual(tileset_info(self.tileset_max, bounds="???"), self.info)


class TilesTest(unittest.TestCase):
    def test_zoom_out_of_bounds(self):
        def should_fail():
            tileset_stub = MockHdf5(
                {"resolutions": {"1": {"values": np.array([[1, 2], [3, 4]])}}}
            )
            tileset_stub.attrs = {}
            tiles(tileset_stub, 2, 0, 0)

        self.assertRaisesRegex(ValueError, r"Zoom level out of bounds", should_fail)

    def test_padding(self):
        tileset = MockHdf5(
            {
                "resolutions": {
                    "1": {
                        "values": np.array([[1.0, 2], [3, 4]])
                        # It's important that there is a float value:
                        # If there isn't, np.nan will be converted
                        # to a large negative integer.
                    }
                }
            }
        )
        tileset.attrs = {}
        zoomed = tiles(tileset, 0, 0, 0)
        self.assertEqual(zoomed.shape, (256, 256))
        assert_array_equal(zoomed[0:2, 0:2], [[1, 2], [3, 4]])
        assert_array_equal(zoomed[2:256, 0], [np.nan for x in range(254)])

    def test_bins(self):
        tileset = MockHdf5(
            {
                "resolutions": {
                    "1": {
                        "values": np.array(
                            [[float(x) for x in range(500)] for y in range(500)]
                        )
                    }
                }
            }
        )
        tileset.attrs = {}

        zoomed_0 = tiles(tileset, 0, 0, 0)
        self.assertEqual(zoomed_0.shape, (256, 256))
        self.assertEqual(zoomed_0[0, 0], 0)

        zoomed_1 = tiles(tileset, 0, 1, 1)
        self.assertEqual(zoomed_1.shape, (256, 256))
        self.assertEqual(zoomed_1[0, 0], 256)
        self.assertEqual(zoomed_1[1, 0], 256)  # Constant dimension
        self.assertEqual(zoomed_1[0, 1], 257)  # Changing dimension
        self.assertEqual(zoomed_1[0, 256 - 13], 499)
        assert_array_equal(zoomed_1[0:1, (256 - 12) : (256 - 12 + 1)], [[np.nan]])
        # Plain assertEqual gave: nan != nan

    def test_zoom(self):
        tileset = MockHdf5(
            {
                "resolutions": {
                    "1": {"values": np.array([[1, 2], [3, 4]])},
                    "5": {"values": np.array([[3, 4], [5, 6]])},
                    "11": {"values": np.array([[5, 6], [7, 8]])},
                }
            }
        )
        tileset.attrs = {}

        zoomed_0 = tiles(tileset, 0, 0, 0)
        assert_array_equal(zoomed_0[0:2, 0:2], [[5, 6], [7, 8]])

        zoomed_1 = tiles(tileset, 1, 0, 0)
        assert_array_equal(zoomed_1[0:2, 0:2], [[3, 4], [5, 6]])

        zoomed_2 = tiles(tileset, 2, 0, 0)
        assert_array_equal(zoomed_2[0:2, 0:2], [[1, 2], [3, 4]])
